"""NGPVAN Activist Code Endpoints"""

from parsons.etl.table import Table
from parsons.ngpvan.utilities import action_parse
import logging

logger = logging.getLogger(__name__)


class ActivistCodes(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_activist_codes(self):
        """
        Get activist codes.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('activistCodes'))
        logger.info(f'Found {tbl.num_rows} activist codes.')
        return tbl

    def get_activist_code(self, activist_code_id):
        """
        Get an activist code.

        `Args:`
            activist_code_id : int
                The activist code id.
        `Returns:`
            dict
                The activist code
        """

        r = self.connection.get_request(f'activistCodes/{activist_code_id}')
        logger.info(f'Found activist code {activist_code_id}.')
        return r

    def toggle_activist_code(self, id, activist_code_id, action, id_type='vanid'):
        # Internal method to apply/remove activist codes. Was previously a public method,
        # but for the sake of simplicity, breaking out into two public methods.

        logger.info(f'Method deprecated. Use apply_activist_code() or remove_activist_code().')

        response = {"activistCodeId": activist_code_id,
                    "action": action_parse(action),
                    "type": "activistCode"}

        r = self.apply_response(id, response, id_type)

        logger.info(f'{id_type.upper()} {id} {action.capitalize()} ' +
                    f'activist code {activist_code_id}')

        return r

    def apply_activist_code(self, id, activist_code_id, id_type='vanid'):
        """
        Apply an activist code to or from a person.

        `Args:`
            id: str
                A valid person id
            activist_code_id: int
                A valid activist code id
            action: str
                Either 'apply' or 'remove'
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
        Returns:
            ``None``
        """

        return self.toggle_activist_code(id, activist_code_id, 'Apply', id_type=id_type)

    def remove_activist_code(self, id, activist_code_id, id_type='vanid'):
        """
        Remove an activist code to or from a person.

        `Args:`
            id: str
                A valid person id
            activist_code_id: int
                A valid activist code id
            action: str
                Either 'apply' or 'remove'
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
        Returns:
            ``None``
        """

        return self.toggle_activist_code(id, activist_code_id, 'Remove', id_type=id_type)



import os


def check(env, field=None):
    """Check if an envrionment variable is set for __init__"""
    if not field:
        try:
            return os.environ[env]
        except KeyError:
            raise KeyError(f'No {env} found. Store as environment variable or '
                           f'pass as an argument.')
    return field



import unittest
import os
from datetime import datetime
import pytz
from parsons.aws.s3 import S3
from parsons.etl.table import Table
import urllib
import time
import shutil
from test.utils import assert_matching_tables

# Requires a s3 credentials stored in aws config or env variable
# to run properly.


@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestS3(unittest.TestCase):

    def setUp(self):

        self.s3 = S3()

        creds = self.s3.aws.session.get_credentials()

        # Create a new bucket
        self.test_bucket = os.environ['S3_TEMP_BUCKET']
        # Trying miss random errors on not finding buckets
        self.s3.create_bucket(self.test_bucket)

        self.test_key = 'test.csv'
        self.tbl = Table([{'first': 'Bob', 'last': 'Smith'}])
        csv_path = self.tbl.to_csv()

        # Sometimes it runs into issues putting the file
        retry = 1

        while retry <= 5:
            try:
                # Put a test file in the bucket
                self.s3.put_file(self.test_bucket, self.test_key, csv_path)
                break
            except:
                print ('Retrying putting file in bucket...')
                retry += 1

    def tearDown(self):

        for k in self.s3.list_keys(self.test_bucket):
            self.s3.remove_file(self.test_bucket, k)

    def test_list_buckets(self):

        # Also tests that create_bucket works (part of setup)

        buckets = self.s3.list_buckets()
        self.assertTrue(self.test_bucket in buckets)

    def test_bucket_exists(self):

        # Test that a bucket that exists returns True
        self.assertTrue(self.s3.bucket_exists(self.test_bucket))

        # Test that a bucket that doesn't exist returns False
        self.assertFalse(self.s3.bucket_exists('idontexist_bucket'))

    def test_list_keys(self):

        # Put a file in the bucket
        csv_path = self.tbl.to_csv()
        key = 'test/test.csv'
        self.s3.put_file(self.test_bucket, key, csv_path)

        # Test that basic bucket list works
        keys = self.s3.list_keys(self.test_bucket, prefix='test/test')
        self.assertTrue(key in keys)

        # Test that prefix filter works -- when there
        keys = self.s3.list_keys(self.test_bucket, prefix='test')
        self.assertTrue(key in keys)

        # Test that prefix filter works -- when not there
        keys = self.s3.list_keys(self.test_bucket, prefix='nope')
        self.assertFalse(key in keys)

    def test_key_exists(self):

        csv_path = self.tbl.to_csv()
        key = 'test/test.csv'
        self.s3.put_file(self.test_bucket, key, csv_path)

        # Test that returns True if key exists
        self.assertTrue(self.s3.key_exists(self.test_bucket, key))

        # Test that returns True if key does not exist
        self.assertFalse(self.s3.key_exists(self.test_bucket, 'akey'))

    def test_list_keys_suffix(self):

        # Put a file in the bucket
        csv_path = self.tbl.to_csv()
        key_1 = 'test/test.csv'
        key_2 = 'test/test.gz'
        self.s3.put_file(self.test_bucket, key_1, csv_path)
        self.s3.put_file(self.test_bucket, key_2, csv_path)

        keys = self.s3.list_keys(self.test_bucket, suffix='csv')
        self.assertTrue(key_1 in keys)
        self.assertFalse(key_2 in keys)

        keys = self.s3.list_keys(self.test_bucket, suffix='gz')
        self.assertFalse(key_1 in keys)
        self.assertTrue(key_2 in keys)

    def test_list_keys_date_modified(self):

        # Set current utc timestamp with timezone
        current_utc = datetime.utcnow().astimezone(pytz.utc)

        # Ensure the files created before now are included
        keys = self.s3.list_keys(self.test_bucket, date_modified_before=current_utc)
        self.assertEqual(len(keys), 1)

        # Ensure the files created after now are not included
        keys = self.s3.list_keys(self.test_bucket, date_modified_after=current_utc)
        self.assertEqual(len(keys), 0)

    def test_put_and_get_file(self):

        # put_file is part of setup, so just testing getting it here

        path = self.s3.get_file(self.test_bucket, self.test_key)
        result_tbl = Table.from_csv(path)
        assert_matching_tables(self.tbl, result_tbl)

    def test_get_url(self):

        # Test that you can download from URL
        url = self.s3.get_url(self.test_bucket, self.test_key)
        csv_table = Table.from_csv(url)

        # Test that the url expires
        url_short = self.s3.get_url(self.test_bucket, self.test_key, expires_in=1)
        time.sleep(2)
        tbl = Table.from_csv(url_short)
        self.assertRaises(urllib.error.HTTPError, print, tbl)

    def test_transfer_bucket(self):

        # Create a destination bucket
        # TODO maybe pull this from an env var as well
        destination_bucket = f"{self.test_bucket}-test"
        dbuk = self.s3.create_bucket(destination_bucket)

        # Transfer
        self.s3.transfer_bucket(self.test_bucket, self.test_key, destination_bucket)

        # Test that object made it
        path = self.s3.get_file(destination_bucket, self.test_key)
        result_tbl = Table.from_csv(path)
        assert_matching_tables(self.tbl, result_tbl)



import unittest
import warnings
import os

from parsons.facebook_ads.facebook_ads import FacebookAds
from parsons.etl.table import Table

users_table = Table([
    {"first": "Bob", "middle": "J", "last": "Smith", "phone": "1234567890", "cell": None,
     "vb_voterbase_dob": "19820413"},
    {"first": "Sue", "middle": "Lucy", "last": "Doe", "phone": None, "cell": "2345678901",
     "vb_voterbase_dob": None},
])

@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestFacebookAdsIntegration(unittest.TestCase):

    def setUp(self):

        self.fb_ads = FacebookAds()

        self.audience_id = self.fb_ads.create_custom_audience(
            name="Test Audience",
            data_source="USER_PROVIDED_ONLY"
            )

    def tearDown(self):
        self.fb_ads.delete_custom_audience(self.audience_id)

    def test_create_custom_audience(self):
        # Audience created in setUp
        self.assertIsNotNone(self.audience_id)

    def test_create_custom_audience_bad_data_source(self):
        self.assertRaises(
            KeyError,
            self.fb_ads.create_custom_audience, name="Something", data_source="INVALID"
            )

    def test_add_users_to_custom_audience(self):
        # Note we don't actually check the results of adding these users, eg. how many were
        # matched by FB. This test just ensures the method doesn't error.
        # TODO See if we can do a more effective test.
        self.fb_ads.add_users_to_custom_audience(self.audience_id, users_table)

    def test_add_users_to_custom_audience_no_valid_columns(self):
        # We don't yet support full names for matching, so this shouldn't work
        tbl = Table([
            {"full name": "Bob Smith"},
            ])
        self.assertRaises(
            KeyError,
            self.fb_ads.add_users_to_custom_audience, self.audience_id, tbl
            )


class TestFacebookAdsUtilities(unittest.TestCase):

    def test_get_match_key_for_column(self):
        # Test just a few of the mappings
        self.assertEqual('EMAIL', FacebookAds._get_match_key_for_column('email'))
        self.assertEqual('EMAIL', FacebookAds._get_match_key_for_column('voterbase_email'))
        self.assertEqual('FN', FacebookAds._get_match_key_for_column('first name'))
        self.assertEqual('FN', FacebookAds._get_match_key_for_column('FIRST-NAME '))
        self.assertEqual('FN', FacebookAds._get_match_key_for_column('vb_tsmart_first_name'))
        self.assertEqual('LN', FacebookAds._get_match_key_for_column('Last Name!'))
        self.assertEqual('ST', FacebookAds._get_match_key_for_column('state code'))
        self.assertEqual('ST', FacebookAds._get_match_key_for_column('vb_vf_source_state'))
        self.assertEqual('GEN', FacebookAds._get_match_key_for_column('vb_voterbase_gender'))
        self.assertEqual('PHONE', FacebookAds._get_match_key_for_column('vb_voterbase_phone_wireless'))
        self.assertIsNone(FacebookAds._get_match_key_for_column('invalid'))

    def test_get_preprocess_key_for_column(self):
        self.assertEqual('DOB YYYYMMDD', FacebookAds._get_preprocess_key_for_column('vb_voterbase_dob'))

    def test_get_match_table_for_users_table(self):
        # This tests basic column matching, as well as the more complex cases like:
        # - Where multiple columns map to the same FB key ("PHONE").
        # - Parsing of a YYYYMMDD birth date column.
        t = FacebookAds.get_match_table_for_users_table(users_table)

        row0 = t[0]
        self.assertEqual(["FN", "LN", "PHONE", "DOBY", "DOBM", "DOBD"], t.columns)
        self.assertEqual("Bob", row0["FN"])
        self.assertEqual("Smith", row0["LN"])
        self.assertEqual("1234567890", row0["PHONE"])
        self.assertEqual("1982", row0["DOBY"])
        self.assertEqual("04", row0["DOBM"])
        self.assertEqual("13", row0["DOBD"])

        row1 = t[1]
        self.assertEqual("Sue", row1["FN"])
        self.assertEqual("Doe", row1["LN"])
        self.assertEqual("2345678901", row1["PHONE"])
        self.assertEqual("", row1["DOBY"])
        self.assertEqual("", row1["DOBM"])
        self.assertEqual("", row1["DOBD"])


    def test_get_match_schema_and_data(self):
        match_table = Table([
            {"FN": "Bob", "LN": "Smith"},
            {"FN": "Sue", "LN": "Doe"},
        ])
        (schema, data) = FacebookAds._get_match_schema_and_data(match_table)
        self.assertEqual(["FN", "LN"], schema)
        self.assertEqual(("Bob", "Smith"), data[0])
        self.assertEqual(("Sue", "Doe"), data[1])




import unittest
import os
import requests_mock
from parsons.mobile_commons.mobile_commons import MobileCommons
from test.utils import validate_list


os.environ['MC_USER'] = 'SOME_USER'
os.environ['MC_PASS'] = 'SOME_PASS'


class TestMobileCommons(unittest.TestCase):

    def setUp(self):

        self.mc = MobileCommons(
            username=os.environ['MC_USER'],
            password=os.environ['MC_PASS'])

    def tearDown(self):

        pass

    def test_clean_dict(self):
        init = [{'@z': [{'@a': 'a'},
                {'@b': 'b'}],
                '@y': 'nothing',
                '@x': {'@l': 'l'}}]

        expected = [{"z": [{"a": "a"},
                    {"b": "b"}],
                    "y": "nothing",
                    "x": {"l": "l"}}]
        self.assertCountEqual(self.mc.connection.clean_dict(init), expected)

    @requests_mock.Mocker()
    def test_campaigns(self, m):

        # Create response
        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <campaigns>'
               '    <campaign id="168866" active="true">'
               '      <name>FWA-SaleforceTestAccount Master Campaign</name>'
               '      <description/>'
               '      <tags>'
               '      </tags>'
               '    </campaign>'
               '    <campaign id="184679" active="true">'
               '      <name>TEST Campagin</name>'
               '      <description>Test campaign created by Daniel Bravo</description>'
               '      <tags>'
               '        <tag>daniel@movementcooperative.org</tag>'
               '      </tags>'
               '    </campaign>'
               '  </campaigns>'
               '</response>')

        m.get(self.mc.connection.uri + 'campaigns', text=xml)

        # Expected Structure
        expected = ['id', 'active', 'name', 'description', 'tags']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.mc.campaigns()))

        # To Do: Test what happens when it doesn't find any Campaigns

    @requests_mock.Mocker()
    def test_campaign(self, m):

        # Initial conditions
        CAMPAIGN_ID = 168866

        # Create response
        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <campaigns>'
               '    <campaign id="168866" active="true">'
               '      <name>FWA-SaleforceTestAccount Master Campaign</name>'
               '      <description/>'
               '      <tags>'
               '      </tags>'
               '    </campaign>'
               '  </campaigns>'
               '</response>')
        m.get(self.mc.connection.uri + 'campaigns', text=xml)

        # Expected Structure
        expected = ['id', 'active', 'name', 'description', 'tags']

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.mc.campaign(CAMPAIGN_ID)))

        # To Do: Test what happens when it doesn't find the Campaign

    @requests_mock.Mocker()
    def test_groups(self, m):

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <groups>'
               '    <group id="15084" type="FilteredGroup" status="active">'
               '      <name>CA California</name>'
               '      <size>4896</size>'
               '    </group>'
               '    <group id="18674" type="UploadedGroup" status="inactive">'
               '      <name>Capps Callers</name>'
               '      <size>90</size>'
               '    </group>'
               '    <group id="20094" type="UploadedGroup" status="inactive">'
               '      <name>Lockyer Callers</name>'
               '      <size>28</size>'
               '    </group>'
               '  </groups>'
               '</response>')

        m.get(self.mc.connection.uri + 'groups', text=xml)

        # Expected Structure
        expected = ['id', 'type', 'status', 'name', 'size']

        self.assertTrue(validate_list(expected, self.mc.groups()))

    @requests_mock.Mocker()
    def test_groups_none(self, m):

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <groups>'
               '  </groups>'
               '</response>')

        m.get(self.mc.connection.uri + 'groups', text=xml)

        # Expects None

        self.assertIsNone(self.mc.groups())

    @requests_mock.Mocker()
    def test_group_members(self, m):

        # Initial conditions
        GROUP_ID = 355559

# FIXXXXX
        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <group name="“WATCH” opt-ins since December 4th" id="355559" type="FilteredGroup">'
               '    <profile id="260832344">'
               '      <first_name/>'
               '      <last_name/>'
               '      <phone_number>19148825677</phone_number>'
               '      <email>rlandstrom11@gmail.com</email>'
               '      <status>Active Subscriber</status>'
               '      <created_at>2017-12-04 21:16:47 UTC</created_at>'
               '      <updated_at>2018-04-30 21:46:07 UTC</updated_at>'
               '      <opted_out_at/>'
               '      <opted_out_source/>'
               '      <source type="mConnect" name="NY_Cuomo_Stop inFRACKstructure_LS1117" id="36653"/>'
               '      <address>'
               '        <street1/>'
               '        <street2/>'
               '        <city>Monroe</city>'
               '        <state>NY</state>'
               '        <postal_code>10950</postal_code>'
               '        <country>US</country>'
               '      </address>'
               '      <location>'
               '        <latitude>41.33065</latitude>'
               '        <longitude>-74.18681</longitude>'
               '        <precision>place</precision>'
               '        <city>Monroe</city>'
               '        <state>NY</state>'
               '        <postal_code>10950</postal_code>'
               '        <country>US</country>'
               '      </location>'
               '      <districts>'
               '        <congressional_district>NY-18</congressional_district>'
               '        <state_upper_district>NY-</state_upper_district>'
               '        <state_lower_district>NY-</state_lower_district>'
               '        <split_district>No</split_district>'
               '      </districts>'
               '      <custom_columns>'
               '        <custom_column name="0715ManyWords"/>'
               '        <custom_column name="0715OneWord"/>'
               '        <custom_column name="ABX_pop_quiz"/>'
               '        <custom_column name="Antibiotic-Resistance_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ1_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ2_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ3_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ4_MJB0614"/>'
               '        <custom_column name="CA_Feb7Bus_MJB0115"/>'
               '        <custom_column name="CON-ID"/>'
               '        <custom_column name="DAPL_Action_TownHall_1_SH"/>'
               '        <custom_column name="DAPL_Action_TownHall_2_SH"/>'
               '        <custom_column name="DAPL_Action_TownHall_3_SH"/>'
               '        <custom_column name="dark_act_petition"/>'
               '        <custom_column name="DC_resist_SH"/>'
               '        <custom_column name="DC_resist2_SH"/>'
               '        <custom_column name="Fossil Fuel_1"/>'
               '        <custom_column name="Fossil Fuel_2"/>'
               '        <custom_column name="Frackdown - Cool? MJB091114"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate2_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate3_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate4_MC0516"/>'
               '        <custom_column name="FWWJOBS"/>'
               '        <custom_column name="general_text"/>'
               '        <custom_column name="Generic_YES-NO"/>'
               '        <custom_column name="Krasner_1"/>'
               '        <custom_column name="Krasner_2"/>'
               '        <custom_column name="MCER_2SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_3SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_4SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_SignUpInfo_SH0516"/>'
               '        <custom_column name="MJB 060712 FR NJ Drive, How many, Need Ride"/>'
               '        <custom_column name="MJB 081613 Obama NY/PA DETAILS"/>'
               '        <custom_column name="MJB 121012 Produce Source"/>'
               '        <custom_column name="MJB 121012 Text Frequency"/>'
               '        <custom_column name="MJB 121012 Water Source"/>'
               '        <custom_column name="MJB042514PeoplesPlatformAgainstFracking"/>'
               '        <custom_column name="MJB070813WatchGaslandPartII"/>'
               '        <custom_column name="NAT_Pruitt1_SH"/>'
               '        <custom_column name="NAT_Pruitt2_SH"/>'
               '        <custom_column name="NAT_Pruitt3_SH"/>'
               '        <custom_column name="NAT_Pruitt4_SH"/>'
               '        <custom_column name="NAT_PublicLandsDay_SH0916"/>'
               '        <custom_column name="NAT_PublicLandsDay2_SH0916"/>'
               '        <custom_column name="NAT_resistance event_SH0217"/>'
               '        <custom_column name="NAT_resistance event2_SH0217"/>'
               '        <custom_column name="NAT_TPP_2ndCall_SH1016"/>'
               '        <custom_column name="NAT_TPP_CALL_SH"/>'
               '        <custom_column name="NAT_TPP_Oct 2016 TPP Activist Call 1_1017SH"/>'
               '        <custom_column name="NAT_TPP_Oct 2016 TPP Activist Call 2_1017SH"/>'
               '        <custom_column name="NJ_Pinelands_Forum_SH0117"/>'
               '        <custom_column name="NJ_Pinelands_Forum2_SH0117"/>'
               '        <custom_column name="NJ_Pinelands Jan 24 Hearing_SH0117"/>'
               '        <custom_column name="NJ_Pinelands Jan 24_SH0117"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_2_SH0916"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_3_SH0916"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_SH0916"/>'
               '        <custom_column name="NTENDrink"/>'
               '        <custom_column name="NY Sheridan Hollow more info 0518"/>'
               '        <custom_column name="OFF_Host Ask_1_SH0417"/>'
               '        <custom_column name="OFF_Host Ask_2_SH0417"/>'
               '        <custom_column name="ombid" created_at="2018-04-30T21:46:07Z" updated_at="2018-04-30T21:46:07Z">'
               '14121        </custom_column>'
               '        <custom_column name="OR Right to Know 091814"/>'
               '        <custom_column name="petition_read_or_sign"/>'
               '        <custom_column name="PLEDGE time of day"/>'
               '        <custom_column name="Resist_address_1"/>'
               '        <custom_column name="Resist_address_2"/>'
               '        <custom_column name="Resist_address_3"/>'
               '        <custom_column name="Resist_address_4"/>'
               '        <custom_column name="REV_Coming_to_Philly"/>'
               '        <custom_column name="Revolution_Opt_in_Coming_Philly"/>'
               '        <custom_column name="Smartphone"/>'
               '        <custom_column name="spanish_english"/>'
               '        <custom_column name="StudentsAct_Interest_SH0916"/>'
               '        <custom_column name="TBTTStudent"/>'
               '        <custom_column name="Top Issue Interest"/>'
               '        <custom_column name="victories_2014election"/>'
               '        <custom_column name="Voted"/>'
               '        <custom_column name="Voter_Registration"/>'
               '        <custom_column name="wolfrally0515"/>'
               '        <custom_column name="WorldWater"/>'
               '        <custom_column name="YES-NO"/>'
               '      </custom_columns>'
               '      <subscriptions>'
               '        <subscription campaign_id="9171" campaign_name="Food &amp; Water Action Master Campaign" campaign_description="" opt_in_path_id="132681" status="Active" opt_in_source="WATCH" created_at="2017-12-04T21:23:58Z" activated_at="2017-12-04T21:23:58Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="12681" campaign_name="FWW General Campaign" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="DRBC call in day (PA, NY, NJ, DE) EP1217 FINAL " created_at="2017-12-05T15:59:30Z" activated_at="2017-12-05T15:59:30Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="82861" campaign_name="New York" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="NY Spectra 12/18 Call Cuomo LS FINAL" created_at="2017-12-18T18:52:37Z" activated_at="2017-12-18T18:52:37Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="89401" campaign_name="Water for the Public" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="NAT Water Privatization Nestle Water Grab LH0318 FINAL" created_at="2018-03-01T23:32:24Z" activated_at="2018-03-01T23:32:24Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="130917" campaign_name="Fundraising" campaign_description="All opt-in paths for people wanting to donate. " opt_in_path_id="" status="Active" opt_in_source="NAT DEV 2018 Sustainer Drive Monsanto survey engagement LH0718 draft copy 9" created_at="2018-08-02T19:34:52Z" activated_at="2018-08-02T19:34:52Z" opted_out_at="" opt_out_source=""/>'
               '      </subscriptions>'
               '      <clicks>'
               '        <click id="114680924">'
               '          <created_at>2018-11-23 22:14:25 UTC</created_at>'
               '          <url>https://www.foodandwaterwatch.org/news/food-water-watch-response-national-climate-assessment-release</url>'
               '          <remote_addr>148.74.84.216</remote_addr>'
               '          <http_referer>148.74.84.216</http_referer>'
               '          <user_agent>Mozilla/5.0 (iPhone; CPU iPhone OS 12_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1</user_agent>'
               '        </click>'
               '        <click id="116297642">'
               '          <created_at>2018-12-04 20:05:18 UTC</created_at>'
               '          <url>https://secure.foodandwaterwatch.org/act/tell-drbc-we-need-full-ban-fracking?ms=onor_mc_12032018_tell-drbc-we-need-full-ban-fracking&amp;oms=onor_mc_12032018_tell-drbc-we-need-full-ban-fracking&amp;phone=19148825677&amp;profile_city=Monroe&amp;profile_email=rlandstrom11%40gmail.com&amp;profile_first_name=&amp;profile_last_name=&amp;profile_postal_code=10950&amp;profile_state=NY</url>'
               '          <remote_addr>216.6.184.136</remote_addr>'
               '          <http_referer>216.6.184.136</http_referer>'
               '          <user_agent>Mozilla/5.0 (iPhone; CPU iPhone OS 12_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Mobile/15E148 Safari/604.1</user_agent>'
               '        </click>'
               '      </clicks>'
               '    </profile>'
               '    <profile id="261988127">'
               '      <first_name/>'
               '      <last_name/>'
               '      <phone_number>19175144109</phone_number>'
               '      <email>rachelmarcohavens@gmail.com</email>'
               '      <status>Active Subscriber</status>'
               '      <created_at>2017-12-11 13:45:10 UTC</created_at>'
               '      <updated_at>2018-04-20 22:22:54 UTC</updated_at>'
               '      <opted_out_at/>'
               '      <opted_out_source/>'
               '      <source type="mConnect" name="NY_Cuomo_Stop inFRACKstructure_LS1117" id="36653"/>'
               '      <address>'
               '        <street1/>'
               '        <street2/>'
               '        <city>Woodstock</city>'
               '        <state>NY</state>'
               '        <postal_code>12498</postal_code>'
               '        <country>US</country>'
               '      </address>'
               '      <location>'
               '        <latitude>42.04136</latitude>'
               '        <longitude>-74.11762</longitude>'
               '        <precision>place</precision>'
               '        <city>Woodstock</city>'
               '        <state>NY</state>'
               '        <postal_code>12498</postal_code>'
               '        <country>US</country>'
               '      </location>'
               '      <districts>'
               '        <congressional_district>NY-19</congressional_district>'
               '        <state_upper_district>NY-</state_upper_district>'
               '        <state_lower_district>NY-</state_lower_district>'
               '        <split_district>No</split_district>'
               '      </districts>'
               '      <custom_columns>'
               '        <custom_column name="0715ManyWords"/>'
               '        <custom_column name="0715OneWord"/>'
               '        <custom_column name="ABX_pop_quiz"/>'
               '        <custom_column name="Antibiotic-Resistance_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ1_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ2_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ3_MJB0614"/>'
               '        <custom_column name="AntibioticsFlowchartQ4_MJB0614"/>'
               '        <custom_column name="CA_Feb7Bus_MJB0115"/>'
               '        <custom_column name="CON-ID"/>'
               '        <custom_column name="DAPL_Action_TownHall_1_SH"/>'
               '        <custom_column name="DAPL_Action_TownHall_2_SH"/>'
               '        <custom_column name="DAPL_Action_TownHall_3_SH"/>'
               '        <custom_column name="dark_act_petition"/>'
               '        <custom_column name="DC_resist_SH"/>'
               '        <custom_column name="DC_resist2_SH"/>'
               '        <custom_column name="Fossil Fuel_1"/>'
               '        <custom_column name="Fossil Fuel_2"/>'
               '        <custom_column name="Frackdown - Cool? MJB091114"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate2_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate3_MC0516"/>'
               '        <custom_column name="Frackopoly_InfoBuyDonate4_MC0516"/>'
               '        <custom_column name="FWWJOBS"/>'
               '        <custom_column name="general_text"/>'
               '        <custom_column name="Generic_YES-NO"/>'
               '        <custom_column name="Krasner_1"/>'
               '        <custom_column name="Krasner_2"/>'
               '        <custom_column name="MCER_2SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_3SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_4SignUpInfo_SH0516"/>'
               '        <custom_column name="MCER_SignUpInfo_SH0516"/>'
               '        <custom_column name="MJB 060712 FR NJ Drive, How many, Need Ride"/>'
               '        <custom_column name="MJB 081613 Obama NY/PA DETAILS"/>'
               '        <custom_column name="MJB 121012 Produce Source"/>'
               '        <custom_column name="MJB 121012 Text Frequency"/>'
               '        <custom_column name="MJB 121012 Water Source"/>'
               '        <custom_column name="MJB042514PeoplesPlatformAgainstFracking"/>'
               '        <custom_column name="MJB070813WatchGaslandPartII"/>'
               '        <custom_column name="NAT_Pruitt1_SH"/>'
               '        <custom_column name="NAT_Pruitt2_SH"/>'
               '        <custom_column name="NAT_Pruitt3_SH"/>'
               '        <custom_column name="NAT_Pruitt4_SH"/>'
               '        <custom_column name="NAT_PublicLandsDay_SH0916"/>'
               '        <custom_column name="NAT_PublicLandsDay2_SH0916"/>'
               '        <custom_column name="NAT_resistance event_SH0217"/>'
               '        <custom_column name="NAT_resistance event2_SH0217"/>'
               '        <custom_column name="NAT_TPP_2ndCall_SH1016"/>'
               '        <custom_column name="NAT_TPP_CALL_SH"/>'
               '        <custom_column name="NAT_TPP_Oct 2016 TPP Activist Call 1_1017SH"/>'
               '        <custom_column name="NAT_TPP_Oct 2016 TPP Activist Call 2_1017SH"/>'
               '        <custom_column name="NJ_Pinelands_Forum_SH0117"/>'
               '        <custom_column name="NJ_Pinelands_Forum2_SH0117"/>'
               '        <custom_column name="NJ_Pinelands Jan 24 Hearing_SH0117"/>'
               '        <custom_column name="NJ_Pinelands Jan 24_SH0117"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_2_SH0916"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_3_SH0916"/>'
               '        <custom_column name="NJ_TeamRetreat_MobileDemo_SH0916"/>'
               '        <custom_column name="NTENDrink"/>'
               '        <custom_column name="NY Sheridan Hollow more info 0518"/>'
               '        <custom_column name="OFF_Host Ask_1_SH0417"/>'
               '        <custom_column name="OFF_Host Ask_2_SH0417"/>'
               '        <custom_column name="ombid" created_at="2018-04-20T22:22:54Z" updated_at="2018-04-20T22:22:54Z">'
               '30471        </custom_column>'
               '        <custom_column name="OR Right to Know 091814"/>'
               '        <custom_column name="petition_read_or_sign"/>'
               '        <custom_column name="PLEDGE time of day"/>'
               '        <custom_column name="Resist_address_1"/>'
               '        <custom_column name="Resist_address_2"/>'
               '        <custom_column name="Resist_address_3"/>'
               '        <custom_column name="Resist_address_4"/>'
               '        <custom_column name="REV_Coming_to_Philly"/>'
               '        <custom_column name="Revolution_Opt_in_Coming_Philly"/>'
               '        <custom_column name="Smartphone"/>'
               '        <custom_column name="spanish_english"/>'
               '        <custom_column name="StudentsAct_Interest_SH0916"/>'
               '        <custom_column name="TBTTStudent"/>'
               '        <custom_column name="Top Issue Interest"/>'
               '        <custom_column name="victories_2014election"/>'
               '        <custom_column name="Voted"/>'
               '        <custom_column name="Voter_Registration"/>'
               '        <custom_column name="wolfrally0515"/>'
               '        <custom_column name="WorldWater"/>'
               '        <custom_column name="YES-NO"/>'
               '      </custom_columns>'
               '      <subscriptions>'
               '        <subscription campaign_id="9171" campaign_name="Food &amp; Water Action Master Campaign" campaign_description="" opt_in_path_id="132681" status="Active" opt_in_source="WATCH" created_at="2017-12-11T13:47:54Z" activated_at="2017-12-11T13:47:54Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="82861" campaign_name="New York" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="NY Spectra 12/18 Call Cuomo LS FINAL" created_at="2017-12-18T18:52:37Z" activated_at="2017-12-18T18:52:37Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="89401" campaign_name="Water for the Public" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="NAT Water Privatization Nestle Water Grab LH0318 FINAL" created_at="2018-03-01T23:28:01Z" activated_at="2018-03-01T23:28:01Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="12681" campaign_name="FWW General Campaign" campaign_description="" opt_in_path_id="" status="Active" opt_in_source="NAT_WATER Act Tell Rep to Co-sponsor_LS0418 Final" created_at="2018-04-25T18:16:58Z" activated_at="2018-04-25T18:16:58Z" opted_out_at="" opt_out_source=""/>'
               '        <subscription campaign_id="130917" campaign_name="Fundraising" campaign_description="All opt-in paths for people wanting to donate. " opt_in_path_id="" status="Active" opt_in_source="NAT DEV 2018 Sustainer Drive Monsanto survey engagement LH0718 draft copy 9" created_at="2018-08-02T19:33:41Z" activated_at="2018-08-02T19:33:41Z" opted_out_at="" opt_out_source=""/>'
               '      </subscriptions>'
               '      <clicks>'
               '        <click id="99490859">'
               '          <created_at>2018-09-07 13:28:49 UTC</created_at>'
               '          <url>https://www.surveymonkey.com/r/VKWSMGF</url>'
               '          <remote_addr>24.161.64.199</remote_addr>'
               '          <http_referer>24.161.64.199</http_referer>'
               '          <user_agent>Mozilla/5.0 (iPhone; CPU iPhone OS 11_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.0 Mobile/15E148 Safari/604.1</user_agent>'
               '        </click>'
               '        <click id="116311775">'
               '          <created_at>2018-12-04 20:41:23 UTC</created_at>'
               '          <url>https://secure.foodandwaterwatch.org/act/tell-drbc-we-need-full-ban-fracking?ms=onor_mc_12032018_tell-drbc-we-need-full-ban-fracking&amp;oms=onor_mc_12032018_tell-drbc-we-need-full-ban-fracking&amp;phone=19175144109&amp;profile_city=Woodstock&amp;profile_email=rachelmarcohavens%40gmail.com&amp;profile_first_name=&amp;profile_last_name=&amp;profile_postal_code=12498&amp;profile_state=NY</url>'
               '          <remote_addr>107.77.223.233</remote_addr>'
               '          <http_referer>107.77.223.233</http_referer>'
               '          <user_agent>Mozilla/5.0 (iPhone; CPU iPhone OS 11_3 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.0 Mobile/15E148 Safari/604.1</user_agent>'
               '        </click>'
               '      </clicks>'
               '    </profile>'
               '  </group>'
               '</response>')

        m.get(self.mc.connection.uri + 'group_members', text=xml)

        # Expected Structure
        expected = ['id', 'first_name', 'last_name', 'phone_number', 'email',
                    'status', 'created_at', 'updated_at', 'opted_out_at',
                    'opted_out_source', 'source', 'address', 'location',
                    'districts', 'custom_columns', 'subscriptions', 'clicks']

        self.assertTrue(
            validate_list(expected, self.mc.group_members(GROUP_ID)))

    @requests_mock.Mocker()
    def test_group_members_group_not_exists(self, m):

        # Initial conditions
        GROUP_ID = 355559

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="false">'
               '  <error id="6" message="Invalid group id"/>'
               '</response>')

        m.get(self.mc.connection.uri + 'group_members', text=xml)

        # Expects None

        self.assertIsNone(self.mc.group_members(GROUP_ID))

    @requests_mock.Mocker()
    def test_group_create(self, m):

        # Initial conditions
        GROUP_NAME = 'This is a TEST group DB'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <group name="This is a TEST group DB" id="436568" type="UploadedGroup">'
               '  </group>'
               '</response>')

        m.get(self.mc.connection.uri + 'create_group', text=xml)

        # Expected Structure
        expected = ['id', 'type', 'name']

        self.assertTrue(
            validate_list(expected, self.mc.group_create(GROUP_NAME)))

    @requests_mock.Mocker()
    def test_group_create_already_exists(self, m):

        # Initial conditions
        GROUP_NAME = 'This is a TEST group DB'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="false">'
               '  <error id="18" message="Invalid name"/>'
               '</response>')

        m.get(self.mc.connection.uri + 'create_group', text=xml)

        # Expects None

        self.assertIsNone(self.mc.group_create(GROUP_NAME))

    @requests_mock.Mocker()
    def test_group_add_members(self, m):

        # Initial conditions
        GROUP_ID = 436568
        PHONES = ['7145555555']

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <group name="This is a TEST group DB" id="436568" type="UploadedGroup">'
               '  </group>'
               '</response>')

        m.get(self.mc.connection.uri + 'add_group_member', text=xml)

        # Expected Structure
        expected = ['id', 'type', 'name']

        self.assertTrue(
            validate_list(
                expected, self.mc.group_add_members(GROUP_ID, PHONES)))

    @requests_mock.Mocker()
    def test_group_add_members_invalid_group(self, m):

        # Initial conditions
        GROUP_ID = 436569
        PHONES = ['7145555555']

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="false">'
               '  <error id="6" message="Invalid group id"/>'
               '</response>')

        m.get(self.mc.connection.uri + 'add_group_member', text=xml)

        # Expects None

        self.assertIsNone(self.mc.group_add_members(GROUP_ID, PHONES))

    @requests_mock.Mocker()
    def test_group_remove_members(self, m):

        # Initial conditions
        GROUP_ID = 436568
        PHONES = ['7145555555']

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <group name="This is a TEST group DB" id="436568" type="UploadedGroup">'
               '  </group>'
               '</response>')

        m.get(self.mc.connection.uri + 'remove_group_member', text=xml)

        # Expected Structure
        expected = ['id', 'type', 'name']

        self.assertTrue(
            validate_list(
                expected, self.mc.group_remove_members(GROUP_ID, PHONES)))

    @requests_mock.Mocker()
    def test_group_remove_members_invalid_group(self, m):

        # Initial conditions
        GROUP_ID = 436569
        PHONES = ['7145555555']

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="false">'
               '  <error id="6" message="Invalid group id"/>'
               '</response>')

        m.get(self.mc.connection.uri + 'remove_group_member', text=xml)

        # Expects None

        self.assertIsNone(self.mc.group_remove_members(GROUP_ID, PHONES))

    @requests_mock.Mocker()
    def test_profiles(self, m):

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <profiles num="2" page="1">'
               '    <profile id="258627572">'
               '      <first_name/>'
               '      <last_name/>'
               '      <phone_number>14435042443</phone_number>'
               '      <email/>'
               '      <status>Profiles with no Subscriptions</status>'
               '      <created_at>2017-11-17 14:47:13 UTC</created_at>'
               '      <updated_at>2017-11-17 14:47:13 UTC</updated_at>'
               '      <opted_out_at/>'
               '      <opted_out_source/>'
               '      <source type="API" name="Ryan Zavislak" email="rzavislak@fwwatch.org"/>'
               '      <address>'
               '        <street1/>'
               '        <street2/>'
               '        <city/>'
               '        <state/>'
               '        <postal_code/>'
               '        <country/>'
               '      </address>'
               '      <location>'
               '        <latitude>39.547557</latitude>'
               '        <longitude>-76.321984</longitude>'
               '        <precision>place</precision>'
               '        <city>Bel Air</city>'
               '        <state>MD</state>'
               '        <postal_code>21014</postal_code>'
               '        <country>US</country>'
               '      </location>'
               '      <districts>'
               '        <congressional_district>MD-1</congressional_district>'
               '        <state_upper_district>MD-</state_upper_district>'
               '        <state_lower_district>MD-</state_lower_district>'
               '        <split_district>No</split_district>'
               '      </districts>'
               '      <custom_columns>'
               '      </custom_columns>'
               '      <integrations>'
               '        <integration constituent_id="" type="Salesforce" synchronized_at=""/>'
               '      </integrations>'
               '      <clicks>'
               '      </clicks>'
               '    </profile>'
               '    <profile id="282126752">'
               '      <first_name>Nance</first_name>'
               '      <last_name>Kidwell</last_name>'
               '      <phone_number>70386981280</phone_number>'
               '      <email>nancekidwell@gmail.com</email>'
               '      <status>Undeliverable</status>'
               '      <created_at>2018-03-02 22:21:42 UTC</created_at>'
               '      <updated_at>2018-03-02 22:21:42 UTC</updated_at>'
               '      <opted_out_at>2018-03-02 22:21:42 UTC</opted_out_at>'
               '      <opted_out_source>Opt-out from API</opted_out_source>'
               '      <source type="API" name="Frakture Bot" email="frakturebot@frakture.com"/>'
               '      <address>'
               '        <street1>1462 Spring Vale Ave</street1>'
               '        <street2/>'
               '        <city>Mc Lean</city>'
               '        <state>VA</state>'
               '        <postal_code>22101-3527</postal_code>'
               '        <country>US</country>'
               '      </address>'
               '      <location>'
               '        <latitude>38.932367</latitude>'
               '        <longitude>-77.185224</longitude>'
               '        <precision>rooftop</precision>'
               '        <city>Mc Lean</city>'
               '        <state>VA</state>'
               '        <postal_code>22101</postal_code>'
               '        <country>US</country>'
               '      </location>'
               '      <districts>'
               '        <congressional_district>VA-10</congressional_district>'
               '        <state_upper_district>VA-32</state_upper_district>'
               '        <state_lower_district>VA-48</state_lower_district>'
               '        <split_district>No</split_district>'
               '      </districts>'
               '      <custom_columns>'
               '      </custom_columns>'
               '      <integrations>'
               '      </integrations>'
               '      <clicks>'
               '      </clicks>'
               '    </profile>'
               '  </profiles>'
               '</response>')

        m.get(self.mc.connection.uri + 'profiles', text=xml)

        # Expected Structure
        expected = ['id', 'first_name', 'last_name', 'phone_number', 'email',
                    'status', 'created_at', 'updated_at', 'opted_out_at',
                    'opted_out_source', 'source', 'address', 'location',
                    'districts', 'custom_columns', 'integrations', 'clicks']

        self.assertTrue(validate_list(expected, self.mc.profiles(limit=2)))

    @requests_mock.Mocker()
    def test_profile_get(self, m):

        # Initial conditions
        PHONE = '7145555555'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <profile id="326709143">'
               '    <first_name>Test</first_name>'
               '    <last_name>User</last_name>'
               '    <phone_number>17145555555</phone_number>'
               '    <email>testuser+7145555555@email.com</email>'
               '    <status>Profiles with no Subscriptions</status>'
               '    <created_at>2018-12-05 01:04:59 UTC</created_at>'
               '    <updated_at>2018-12-05 01:04:59 UTC</updated_at>'
               '    <opted_out_at/>'
               '    <opted_out_source/>'
               '    <source type="API" name="Daniel Bravo" email="daniel@movementcooperative.org"/>'
               '    <address>'
               '      <street1/>'
               '      <street2/>'
               '      <city/>'
               '      <state/>'
               '      <postal_code/>'
               '      <country>US</country>'
               '    </address>'
               '    <location>'
               '      <latitude>33.844983</latitude>'
               '      <longitude>-117.952151</longitude>'
               '      <precision>place</precision>'
               '      <city>Anaheim</city>'
               '      <state>CA</state>'
               '      <postal_code>92801</postal_code>'
               '      <country>US</country>'
               '    </location>'
               '    <districts>'
               '      <congressional_district>CA-</congressional_district>'
               '      <state_upper_district>CA-</state_upper_district>'
               '      <state_lower_district>CA-</state_lower_district>'
               '      <split_district>No</split_district>'
               '    </districts>'
               '    <custom_columns>'
               '    </custom_columns>'
               '    <integrations>'
               '    </integrations>'
               '    <clicks>'
               '    </clicks>'
               '  </profile>'
               '</response>')

        m.get(self.mc.connection.uri + 'profile', text=xml)

        # Expected Structure
        expected = ['id', 'first_name', 'last_name', 'phone_number', 'email',
                    'status', 'created_at', 'updated_at', 'opted_out_at',
                    'opted_out_source', 'source', 'address', 'location',
                    'districts', 'custom_columns', 'integrations', 'clicks']

        self.assertTrue(validate_list(expected, self.mc.profile_get(PHONE)))

    @requests_mock.Mocker()
    def test_profile_get_invalid_phone(self, m):

        # Initial conditions
        PHONE = '71455555558'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="false">'
               '  <error id="5" message="Invalid phone number"/>'
               '</response>')

        m.get(self.mc.connection.uri + 'profile', text=xml)

        # Expects None

        self.assertIsNone(self.mc.profile_get(PHONE))

    @requests_mock.Mocker()
    def test_profile_update(self, m):

        # Initial conditions
        PHONE = '7145555555'
        NAME = 'Testupdate'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <profile id="326709143">'
               '    <first_name>Testupdate</first_name>'
               '    <last_name>User</last_name>'
               '    <phone_number>17145555555</phone_number>'
               '    <email>testuser+7145555555@email.com</email>'
               '    <status>Profiles with no Subscriptions</status>'
               '    <created_at>2018-12-05 01:04:59 UTC</created_at>'
               '    <updated_at>2018-12-05 09:00:48 UTC</updated_at>'
               '    <opted_out_at/>'
               '    <opted_out_source/>'
               '    <source type="API" name="Daniel Bravo" email="daniel@movementcooperative.org"/>'
               '    <address>'
               '      <street1/>'
               '      <street2/>'
               '      <city/>'
               '      <state/>'
               '      <postal_code/>'
               '      <country>US</country>'
               '    </address>'
               '    <location>'
               '      <latitude>33.844983</latitude>'
               '      <longitude>-117.952151</longitude>'
               '      <precision>place</precision>'
               '      <city>Anaheim</city>'
               '      <state>CA</state>'
               '      <postal_code>92801</postal_code>'
               '      <country>US</country>'
               '    </location>'
               '    <districts>'
               '      <congressional_district>CA-</congressional_district>'
               '      <state_upper_district>CA-</state_upper_district>'
               '      <state_lower_district>CA-</state_lower_district>'
               '      <split_district>No</split_district>'
               '    </districts>'
               '    <custom_columns>'
               '    </custom_columns>'
               '    <integrations>'
               '    </integrations>'
               '    <clicks>'
               '    </clicks>'
               '  </profile>'
               '</response>')

        m.post(self.mc.connection.uri + 'profile_update', text=xml)

        # Expected Structure
        expected = ['id', 'first_name', 'last_name', 'phone_number', 'email',
                    'status', 'created_at', 'updated_at', 'opted_out_at',
                    'opted_out_source', 'source', 'address', 'location',
                    'districts', 'custom_columns', 'integrations', 'clicks']

        self.assertTrue(
            validate_list(
                expected, self.mc.profile_update(PHONE, first_name=NAME)))

    @requests_mock.Mocker()
    def test_profile_opt_out(self, m):

        # Initial conditions
        PHONE = '7145555555'

        xml = ('<?xml version="1.0" encoding="UTF-8"?>'
               '<response success="true">'
               '  <profile id="301650791">'
               '    <first_name/>'
               '    <last_name/>'
               '    <phone_number>17142706998</phone_number>'
               '    <email/>'
               '    <status>Undeliverable</status>'
               '    <created_at>2018-07-12 19:48:36 UTC</created_at>'
               '    <updated_at>2018-12-05 09:34:51 UTC</updated_at>'
               '    <opted_out_at>2018-12-05 09:34:51 UTC</opted_out_at>'
               '    <opted_out_source>Opt-out from API</opted_out_source>'
               '    <source type="API" name="Daniel Bravo" email="daniel@movementcooperative.org"/>'
               '    <address>'
               '      <street1/>'
               '      <street2/>'
               '      <city/>'
               '      <state/>'
               '      <postal_code/>'
               '      <country>US</country>'
               '    </address>'
               '    <location>'
               '      <latitude>33.844983</latitude>'
               '      <longitude>-117.952151</longitude>'
               '      <precision>place</precision>'
               '      <city>Anaheim</city>'
               '      <state>CA</state>'
               '      <postal_code>92801</postal_code>'
               '      <country>US</country>'
               '    </location>'
               '    <districts>'
               '      <congressional_district>CA-</congressional_district>'
               '      <state_upper_district>CA-</state_upper_district>'
               '      <state_lower_district>CA-</state_lower_district>'
               '      <split_district>No</split_district>'
               '    </districts>'
               '    <custom_columns>'
               '    </custom_columns>'
               '    <subscriptions>'
               '      <subscription campaign_id="184679" campaign_name="TEST Campagin" campaign_description="Test campaign created by Daniel Bravo" opt_in_path_id="272075" status="Opted-Out" opt_in_source="Daniel Bravo" created_at="2018-12-05T09:22:27Z" activated_at="2018-12-05T09:22:27Z" opted_out_at="2018-12-05T09:34:51Z" opt_out_source="Opt-out from API"/>'
               '    </subscriptions>'
               '    <integrations>'
               '    </integrations>'
               '    <clicks>'
               '    </clicks>'
               '  </profile>'
               '</response>')

        m.post(self.mc.connection.uri + 'profile_opt_out', text=xml)

        # Expected Structure
        expected = ['id', 'first_name', 'last_name', 'phone_number', 'email',
                    'status', 'created_at', 'updated_at', 'opted_out_at',
                    'opted_out_source', 'source', 'address', 'location',
                    'districts', 'custom_columns', 'integrations', 'clicks',
                    'subscriptions']

        self.assertTrue(
            validate_list(
                expected, self.mc.profile_opt_out(PHONE)))

if __name__ == "__main__":
    unittest.main()



import os
import pytest

"""
Use this as a marker before any tests that hit live services. That way they'll only run
if you set the "LIVE_TEST" env var.
Example usage:

@mark_live_test
def test_something():
    service = SomeService()
    ...
"""
mark_live_test = pytest.mark.skipif(not os.environ.get('LIVE_TEST'),
    reason='Skipping because not running live test')

# Tests whether a table has the expected structure
def validate_list(expected_keys, table):

    if set(expected_keys) != set(table.columns):
        raise KeyError('Not all expected keys found.')

    return True

def assert_matching_tables(table1, table2, ignore_headers=False):
    if ignore_headers:
        data1 = table1.data
        data2 = table2.data
    else:
        data1 = table1
        data2 = table2

    for r1, r2 in zip(data1, data2):
        # Cast both rows to lists, in case they are different types of collections
        assert list(r1) == list(r2)



# Adapted from Gmail API tutorial https://developers.google.com/gmail/api
from apiclient import errors
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.application import MIMEApplication
from email.encoders import encode_base64
from email.utils import parseaddr
from googleapiclient.discovery import build
from httplib2 import Http
from oauth2client import file, client, tools
from validate_email import validate_email
import base64
import logging
import mimetypes
import os

# BUG: can't send files equal to or larger than 6MB
# There is a possible fix
# `BUG-1` are the changes that would need to be made

# BUG-1
# import httplib2shim
# import sys

logger = logging.getLogger(__name__)

SCOPES = 'https://www.googleapis.com/auth/gmail.send'


class Gmail(object):
    """Create a Gmail object, for sending emails.

    `Args:`
        creds_path: str
            The path to the credentials.json file.
        token_path: str
            The path to the token.json file.
    """

    def __init__(self, creds_path=None, token_path=None):

        self.user_id = 'me'

        if not creds_path:
            raise ValueError("Invalid path to credentials.json.")

        if not token_path:
            raise ValueError("Invalid path to token.json.")

        self.store = file.Storage(token_path)
        self.creds = self.store.get()

        # BUG-1
        # http = httplib2shim.Http()

        if not self.creds or self.creds.invalid:
            flow = client.flow_from_clientsecrets(creds_path, SCOPES)
            self.creds = tools.run_flow(flow, self.store)

            # BUG-1
            # self.creds = self.run_flow(flow, self.store, http=http)

        self.service = build('gmail', 'v1', http=self.creds.authorize(Http()))

        # BUG-1
        # self.service = build('gmail', 'v1', http=self.creds.authorize(http))

    def create_message_simple(self, sender, to, subject, message_text):
        """Create a text-only message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
        `Returns:`
            dict
                An object containing a base64url encoded email object.
        """
        logger.info("Creating a simple message...")

        message = MIMEText(message_text)
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        msg = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

        logger.debug(msg)
        logger.info("Encoded simple message sucessfully created.")

        return msg

    def create_message_html(self, sender, to, subject, message_text,
                            message_html):
        """Create an html message for an email.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address(es) of the recipient(s).
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message.
        `Returns:`
            dict
                An object containing a base64url encoded email object.
        """
        logger.info("Creating an html message...")

        message = MIMEMultipart('alternative')
        message['subject'] = subject
        message['from'] = sender
        message['to'] = to
        if message_text:
            message.attach(MIMEText(message_text, 'plain'))
        message.attach(MIMEText(message_html, 'html'))

        msg = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

        logger.debug(msg)
        logger.info("Encoded html message sucessfully created.")

        return msg

    def create_message_attachments(self, sender, to, subject, message_text,
                                   files, message_html=None):
        """Create a message for an email that includes an attachment.

        `Args:`
            sender: str
                Email address of the sender.
            to: str
                Email address of the receiver.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            files: list
                The path(s) to the file(s) to be attached.
            message_html: str
                Optional; The html formatted text of the email message.
        `Returns:`
            dict
                An object containing a base64url encoded email object.
        """
        logger.info("Creating a message with attachments...")

        message = MIMEMultipart('alternative')
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        msg = MIMEText(message_text, 'plain')
        message.attach(msg)

        if message_html:
            html = MIMEText(message_html, 'html')
            message.attach(html)

        for f in files:
            content_type, encoding = mimetypes.guess_type(f)
            logger.debug(
                f"(File: {f}, Content-type: {content_type}, "
                f"Encoding: {encoding})")

            if content_type is None or encoding is not None:
                content_type = 'application/octet-stream'

            main_type, sub_type = content_type.split('/', 1)

            if main_type == 'text':
                logger.info("Added a text file.")
                fp = open(f, 'rb')
                msg = MIMEText(fp.read(), _subtype=sub_type, _charset='utf-8')
                fp.close()

            elif main_type == 'image':
                logger.info("Added an image file.")
                fp = open(f, 'rb')
                msg = MIMEImage(fp.read(), _subtype=sub_type)
                filename = os.path.basename(f)
                msg.add_header('Content-ID', f'<{filename}>')
                fp.close()

            elif main_type == 'audio':
                logger.info("Added an audio file.")
                fp = open(f, 'rb')
                msg = MIMEAudio(fp.read(), _subtype=sub_type)
                fp.close()

            elif main_type == 'application':
                logger.info("Added an application file.")
                fp = open(f, 'rb')
                msg = MIMEApplication(fp.read(), _subtype=sub_type)
                fp.close()

            else:
                logger.info("Added an unknown-type file.")
                fp = open(f, 'rb')
                msg = MIMEBase(main_type, sub_type)
                msg.set_payload(fp.read())
                encode_base64(msg)
                fp.close()

            filename = os.path.basename(f)
            msg.add_header(
                'Content-Disposition', 'attachment', filename=filename)
            message.attach(msg)

        m = {'raw': base64.urlsafe_b64encode(message.as_bytes()).decode()}

        logger.debug(m)
        logger.info("Encoded message with attachments sucessfully created.")

        return m

    def _validate_email_string(self, str):
        logger.debug(f"Validating email {str}...")
        realname, email_addr = parseaddr(str)

        if not email_addr:
            raise ValueError("Invalid email address.")

        if not validate_email(email_addr):
            raise ValueError("Invalid email address.")

        return True

    def send_message(self, message, user_id=None):
        """Send an email message.

        `Args:`
            message: dict
                Message to be sent as a base64url encode object.
            user_id: str
                Optional; User's email address. Defaults to the special value
                "me" which is used to indicate the authenticated user.
        `Returns:`
            dict
                A Users.messages object see `https://developers.google.com/gmail/api/v1/reference/users/messages#resource.` # noqa
                for more info.
        """
        logger.info("Sending a message...")
        if not user_id:
            user_id = self.user_id
        try:
            message = (self.service.users().messages()
                       .send(userId=user_id, body=message).execute())
        except errors.HttpError:
            logger.exception(
                'An error occurred: while attempting to send a message.')
            raise
        else:
            logger.debug(message)
            logger.info(
                f"Message sent succesfully (Message Id: {message['id']})")

            return message

    def send_email(self, sender, to, subject, message_text, message_html=None,
                   files=None):
        """Send an email message.

        `Args:`
            sender: str
                Email address of the sender.
            to: str or list
                Email address(es) of the receiver(s). Must be in correct email
                string syntax. For example, `name@email.com` or
                `"Name" <email@email.com>`.
            subject: str
                The subject of the email message.
            message_text: str
                The text of the email message.
            message_html: str
                The html formatted text of the email message. If ommitted, the
                email is sent a text-only body.
            files: str or list
                The path to the file(s) to be attached.

        `Returns:`
            An object containing a base64url encoded email object.
        """
        logger.info("Preparing to send and email...")

        logger.info("Validating email(s)")
        if isinstance(to, list):
            if len(to) == 0:
                raise EmptyListError("Must contain at least 1 email.")

            for e in to:
                self._validate_email_string(e)

            to = ', '.join(to)

        elif isinstance(to, str):
            self._validate_email_string(to)

        if not message_html and not files:
            msg_type = 'simple'
            msg = self.create_message_simple(sender, to, subject, message_text)

        elif not files:
            msg_type = 'html'
            msg = self.create_message_html(
                sender, to, subject, message_text, message_html)
        else:
            msg_type = 'attachments'
            if isinstance(files, str):
                files = [files]

            msg = self.create_message_attachments(
                sender, to, subject, message_text, files, message_html)

        logger.info(f"Sending a(n) {msg_type} email...")

        self.send_message(msg)

        logger.info("Email sent succesfully.")


class EmptyListError(IndexError):
    """Throw when a list is empty that should contain at least 1 element."""
    pass



"""Mobile Commons Profiles Endpoints."""


class Profiles(object):
    """Class for profiles endpoints."""

    def __init__(self, mc_connection):
        """Initialize the Profiles class.

        `Args:`
            mc_connection: MobileCommonsConnector
                The connector to access the Mobile Commons API.
        """
        self.connection = mc_connection

    def profiles(self, phone_number=None, from_date=None, to_date=None,
                 limit=None, page=None, include_custom_columns=None,
                 include_subscriptions=None, include_clicks=None,
                 include_members=None):
        """Return a list of profiles.

        `Args:`
            phone_number: list
                Optional; Limits the returned profiles matching the provided
                phone numbers. Phone numbers should be specified with country
                code.
            from_date: str
                Optional; Limits the returned profiles to ones updated after
                or on this date time. ISO-8601 format
            to_date: str
                Optional; Limits the returned profiles to ones updated before
                or on this date time. ISO-8601 forma
            limit: int
                Optional; Limits the number of returned profiles.
                Maximum of 1000
            page: int
                Optional; Specifies which page, of the total number of pages
                of results, to return
            include_custom_columns: boolean
                Optional; *Optional* default 'true' - allows exclusion of custom
                columns associated with profiles, pass 'false' to limit
            include_subscriptions: boolean
                Optional; *Optional* default 'true' - allows exclusion of
                subscriptions for each profile, pass 'false' to limit
            include_clicks: boolean
                Optional; *Optional* default 'true' - allows exclusion of clicks
            include_members: boolean
                Optional; *Optional* default 'true' - allows exclusion of
                profile member records maintained for integrations
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'profiles'

        if phone_number:
            phone_number = ','.join(phone_number)

        args = {'phone_number': phone_number,
                'from': from_date,
                'to': to_date,
                'limit': limit,
                'page': page,
                'include_custom_columns': include_custom_columns,
                'include_subscriptions': include_subscriptions,
                'include_clicks': include_clicks,
                'include_members': include_members}

        response = self.connection.request_paginate(
            url, 'profiles', args=args, resp_type='xml')

        if response:
            return self.connection.output(response)
        else:
            return None

    def profile_get(self, phone_number, company=None, include_messages=False,
                    include_donations=False):
        """Return a single profile record.

        `Args:`
            phone_number: str
                Required; The phone number for the profile to return.
            company: str
                Optional; If different that the one specified for the
                connection. Default is the firm.
            include_messages: boolean
                Optional; Set to true to include associated text messages.
                Default is false.
            include_donations: boolean
                Optional; Set to true to include associated mobile giving
                donations, if any. Default is false.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'profile'

        args = {'phone_number': phone_number,
                'company': company,
                'include_messages': include_messages,
                'include_donations': include_donations}

        response = self.connection.request(url, args=args, resp_type='xml')

        if response['response']['success'] == 'true':
            return self.connection.output(response['response']['profile'])
        else:
            return None

    def profile_update(self, phone_number, email=None, postal_code=None,
                       first_name=None, last_name=None, street1=None,
                       street2=None, city=None, state=None, country=None,
                       custom_fields={}, opt_in_path_id=None):
        """Create or update a profile.

        `Args:`
            phone_number: str
                Required; The phone number for the profile to update.
            email: str
                Optional; New email for the profile.
            postal_code: str
                Optional; New postal code for the profile.
            first_name: str
                Optional; New firstname for the profile.
            last_name: str
                Optional; New lastname for the profile.
            street1: str
                Optional; New street1 for the profile.
            street2: str
                Optional; New street2 for the profile.
            city: str
                Optional; New city for the profile.
            state: str
                Optional; New state for the profile.
            country: str
                Optional; New country for the profile.
            custom_fields: dict
                Optional; A dict of custom fields and their new values for the
                profile.
            opt_in_path_id: str
                Optional; New opt_in_path_id for the profile.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'profile_update'

        post_data = {
            'phone_number': phone_number,
            'email': email,
            'postal_code': postal_code,
            'first_name': first_name,
            'last_name': last_name,
            'street1': street1,
            'street2': street2,
            'city': city,
            'state': state,
            'country': country,
            'opt_in_path_id': opt_in_path_id}

        # Add each custom field as its own data point
        post_data.update(custom_fields)

        response = self.connection.request(
            url, req_type='POST', post_data=post_data, resp_type='xml')

        if response['response']['success'] == 'true':
            return self.connection.output(response['response']['profile'])
        else:
            return None

    def profile_opt_out(self, phone_number, campaign_id=None,
                        subscription_id=None):
        """Opt out a profile from a campaign, subscription or all.

        `Args:`
            phone_number: str
                Required; The phone number for the profile to opt out.
            campaign_id: int
                Optional; Opt-out this campaign only. Default is all campaigns.
            subscription_id:int
                Optional; Opt-out this subscription only. Default is all
                subscriptions.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        url = self.connection.uri + 'profile_opt_out'

        post_data = {'phone_number': phone_number,
                     'campaign_id': campaign_id,
                     'subscription_id': subscription_id}

        response = self.connection.request(
            url, req_type='POST', post_data=post_data, resp_type='xml')

        if response['response']['success'] == 'true':
            return self.connection.output(response['response']['profile'])
        else:
            return None




def arg_format(arg):
    """
    Many APIs require arguments to formatted like this 'thisTypeConfig' which is not the standard
    for python so this method takes an argument 'this_type_config' and returns it as
    'thisTypeConfig'
    """

    arg_list = arg.split('_')
    arg_list = [a.capitalize() for a in arg_list]
    arg_list[0] = arg_list[0].lower()

    return ''.join(arg_list)


def remove_empty_keys(dirty_dict):
    """
    Remove empty keys from a dictionary. This method is useful when passing jsons
    in which a null field will update the value to null and you don't want that.
    """

    clean_dict = {}

    for k, v in dirty_dict.items():
        if v:
            clean_dict[k] = v

    return clean_dict



from requests import request
import math
import json
import time
from parsons import Table
from parsons.utilities import check_env
import logging

logger = logging.getLogger(__name__)

COPPER_URI = "https://api.prosperworks.com/developer_api/v1"


class Copper(object):
    """
    Instantiate Copper Class

    `Args:`
        user_email:
            The email of the API user for Copper. Not required if ``COPPER_USER_EMAIL``
            env variable set.
        api_key:
            The Copper provided application key. Not required if ``COPPER_API_KEY``
            env. variable set.
    `Returns:`
        Copper Class
    """

    def __init__(self, user_email=None, api_key=None):

        self.api_key = check_env.check('COPPER_API_KEY', api_key)
        self.user_email = check_env.check('COPPER_USER_EMAIL', user_email)
        self.uri = COPPER_URI

    def base_request(self, endpoint, req_type, page=1, page_size=200, filters=None):
        # Internal Request Method

        url = self.uri + endpoint

        # Authentication must be done through headers, requests HTTPBasicAuth doesn't work
        headers = {
            'X-PW-AccessToken': self.api_key,
            'X-PW-Application': "developer_api",
            'X-PW-UserEmail': self.user_email,
            'Content-Type': "application/json"
        }

        payload = {}
        if filters is not None:
            if len(filters) > 0 and isinstance(filters, dict):
                payload.update(filters)

        # GET request with non-None data arg is malformed
        if req_type == 'GET':
            return request(req_type, url, params=json.dumps(payload), headers=headers)
        else:
            payload["page_number"] = page
            payload["page_size"] = page_size

            return request(req_type, url, data=json.dumps(payload), headers=headers)

    def paginate_request(self, endpoint, req_type, page_size=200, filters=None):
        # Internal pagination method

        page = 1
        total_pages = 2
        blob = []
        only_page = False

        if isinstance(filters, dict):
            # Assume user wants just that page if page_number specified in filters
            if 'page_number' in filters:
                page = filters['page_number']
                # Ensure exactly one loop
                total_pages = page
                rows = f'{str(page_size)} or less'
                only_page = True
        else:
            filters = {}

        while page <= total_pages:

            r = self.base_request(
                endpoint,
                req_type,
                page_size=page_size,
                page=page,
                filters=filters
            )

            if page == 1:
                if 'X-Pw-Total' in r.headers and not only_page:
                    rows = r.headers['X-Pw-Total']
                    total_pages = int(math.ceil(int(rows)/float(page_size)))
                else:
                    rows = f'{str(page_size)} or less'
                    total_pages = 1
            logger.info(f"Retrieving page {page} of {total_pages}, total rows: {rows}")
            page += 1

            if r.text == "":
                return []
            # Avoid too many layers of nesting if possible
            if isinstance(json.loads(r.text), list):
                blob.extend(json.loads(r.text))
            else:
                blob.append(json.loads(r.text))
            # Wait for 1 second to avoid hitting rate limits
            time.sleep(1)

        return blob

    def get_people(self, filters=None, tidy=False):
        """
        Get people

        `Args:`
            `filters: dict`
                Optional; pass additional parameters to filter the records returned.
                See `Copper documentation <https://developer.copper.com/?version=latest#9c15869b-c894-4fa2-9346-d65a6602c129>`_ for choices
            `tidy: boolean or int`
                Optional; unpack list and dict columns as additional rows instead of columns
                If `True`: creates new table out of unpacked rows
                If 'int': adds rows to original table if max rows per key <= given number
                (so `tidy=0` guarantees new table)

        `Returns:`
            List of dicts of Parsons Tables:
                * people
                * people_emails
                * people_phone_numbers
                * people_custom_fields
                * people_socials
                * people_websites
        """ # noqa: E501,E261

        return self.get_standard_object("people", filters=filters, tidy=tidy)

    def get_companies(self, filters=None, tidy=False):
        """
        Get companies

        `Args:`
            `filters: dict`
                Optional; pass additional parameters to filter the records returned.
                See `Copper documentation <https://developer.copper.com/?version=latest#0b4f267f-3180-4041-861c-13f3cf17bcf9>`_ for choices
            `tidy: boolean or int`
                Optional; unpack list and dict columns as additional rows instead of columns
                If `True`: creates new table out of unpacked rows
                If 'int': adds rows to original table if max rows per key <= given number
                (so `tidy=0` guarantees new table)

        `Returns:`
            List of dicts of Parsons Tables:
                * companies
                * companies_phone_numbers
                * companies_custom_fields
                * companies_socials
                * companies_websites
        """ # noqa: E501,E261

        return self.get_standard_object("companies", filters=filters, tidy=tidy)

    def get_activities(self, filters=None, tidy=False):
        """
        Get activities

        `Args:`
            `filters: dict`
                Optional; pass additional parameters to filter the records returned.
                See `Copper documentation <https://developer.copper.com/?version=latest#d2e6ddd8-6699-4ff3-87e3-1febb0410dc9>`_ for choices
                Optional; unpack list and dict columns as additional rows instead of columns
                If `True`: creates new table out of unpacked rows
                If 'int': adds rows to original table if max rows per key <= given number
                (so `tidy=0` guarantees new table)

        `Returns:`
            List of dicts of Parsons Tables:
                * activities
        """ # noqa: E501,E261

        return self.get_standard_object("activities", filters=filters, tidy=tidy)

    def get_opportunities(self, filters=None, tidy=False):
        """
        Get opportunities (i.e. donations)

        `Args:`
            `filters: dict`
                Optional; pass additional parameters to filter the records returned.
                See `Copper documentation <https://developer.copper.com/?version=latest#5bb8adc1-137f-46bf-aa86-7df037840e57>`_ for choices
                Optional; unpack list and dict columns as additional rows instead of columns
                If `True`: creates new table out of unpacked rows
                If 'int': adds rows to original table if max rows per key <= given number
                (so `tidy=0` guarantees new table)

        `Returns:`
            List of dicts of Parsons Tables:
                * opportunities
                * opportunities_custom_fields
        """ # noqa: E501,E261

        return self.get_standard_object("opportunities", filters=filters, tidy=tidy)

    def get_standard_object(self, object_name, filters=None, tidy=False):
        # Retrieve and process a standard endpoint object (e.g. people, companies, etc.)

        logger.info(f"Retrieving {object_name} records.")
        blob = self.paginate_request(f"/{object_name}/search", req_type='POST', filters=filters)

        return self.process_json(blob, object_name, tidy=tidy)

    def get_custom_fields(self):
        """
        Get custom fields

        `Args:`
            `filters: dict`
            Optional; pass additional parameters to filter the records returned.
            See `Copper documentation <https://developer.copper.com/?version=latest#bf389290-0c19-46a7-85bf-f5e6884fa4e1>`_ for choices

        `Returns:`
            List of dicts of Parsons Tables:
                * custom_fields
                * custom_fields_available
                * custom_fields_options
        """ # noqa: E501,E261

        logger.info("Retrieving custom fields.")
        blob = self.paginate_request('/custom_field_definitions/', req_type='GET')
        return self.process_custom_fields(blob)

    def get_activity_types(self):
        """
        Get activity types

        `Args:`
            `filters: dict`
            Optional; pass additional parameters to filter the records returned.
            See `Copper documentation <https://developer.copper.com/?version=latest#6bd339f1-f0de-48b4-8c34-5a5e245e036f>`_ for choices

        `Returns:`
            List of dicts of Parsons Tables:
                * activitiy_types
        """ # noqa: E501,E261

        logger.info("Retrieving activity types.")

        response = self.paginate_request('/activity_types/', req_type='GET')
        orig_table = Table(response)
        at_user = orig_table.long_table([], 'user', prepend=False)
        at_sys = orig_table.long_table([], 'system', prepend=False)
        Table.concat(at_sys, at_user)

        return [{'name': 'activity_types', 'tbl': at_sys}]

    def get_contact_types(self):
        """
        Get contact types

        `Args:`
            `filters: dict`
            Optional; pass additional parameters to filter the records returned.
            See `Copper documentation <https://developer.copper.com/?version=latest#8b6e6ed8-c594-4eed-a2af-586aa2100f09>`_ for choices

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """ # noqa: E501,E261

        response = self.paginate_request('/contact_types/', req_type='GET')
        return Table(response)

    def process_json(self, json_blob, obj_type, tidy=False):
        # Internal method for converting most types of json responses into a list of Parsons tables

        # Output goes here
        table_list = []

        # Original table & columns
        obj_table = Table(json_blob)
        cols = obj_table.get_columns_type_stats()
        list_cols = [x['name'] for x in cols if 'list' in x['type']]
        dict_cols = [x['name'] for x in cols if 'dict' in x['type']]

        # Unpack all list columns
        if len(list_cols) > 0:
            for l in list_cols:
                # Check for nested data
                list_rows = obj_table.select_rows(
                    lambda row: isinstance(row[l], list)
                    and any(isinstance(x, dict) for x in row[l])
                )
                # Add separate long table for each column with nested data
                if list_rows.num_rows > 0:
                    logger.debug(l, 'is a nested column')
                    if len([x for x in cols if x['name'] == l]) == 1:
                        table_list.append({
                            'name': f'{obj_type}_{l}',
                            'tbl': obj_table.long_table(['id'], l)
                        })
                    else:
                        # Ignore if column doesn't exist (or has multiples)
                        continue
                else:
                    if tidy is False:
                        logger.debug(l, 'is a normal list column')
                        obj_table.unpack_list(l)

        # Unpack all dict columns
        if len(dict_cols) > 0 and tidy is False:
            for d in dict_cols:
                logger.debug(d, 'is a dict column')
                obj_table.unpack_dict(d)

        if tidy is not False:
            packed_cols = list_cols + dict_cols
            for p in packed_cols:
                if p in obj_table.columns:
                    logger.debug(p, 'needs to be unpacked into rows')

                    # Determine whether or not to expand based on tidy
                    unpacked_tidy = obj_table.unpack_nested_columns_as_rows(p, expand_original=tidy)
                    # Check if column was removed as sign it was unpacked into separate table
                    if p not in obj_table.columns:
                        table_list.append({
                            'name': f'{obj_type}_{p}',
                            'tbl': unpacked_tidy
                        })
                    else:
                        obj_table = unpacked_tidy

        # Original table will have had all nested columns removed
        if len(obj_table.columns) > 1:
            table_list.append({'name': obj_type, 'tbl': obj_table})

        return table_list

    def process_custom_fields(self, json_blob):
        # Internal method to convert custom fields responses into a list of Parsons tables

        # Original table & columns
        custom_fields = Table(json_blob)

        # Available On
        available_on = custom_fields.long_table(['id'], 'available_on')

        # Options
        options = custom_fields.long_table(['id', 'name'], 'options')

        return [{'name': 'custom_fields', 'tbl': custom_fields},
                {'name': 'custom_fields_available', 'tbl': available_on},
                {'name': 'custom_fields_options', 'tbl': options}]



# Dictionaries used to map column names to valid argument names

LEAD_COLUMN_MAP = {'first_name': ['first', 'fn', 'firstname'],
                   'last_name': ['last', 'ln', 'lastname'],
                   'phone_number': ['phone', 'cell', 'phonenumber'],
                   'email': ['email_address', 'emailaddress'],
                   'follow_up': ['followup']
                   }



from parsons.utilities import json_format
import logging

logger = logging.getLogger(__name__)


class People(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def find_person(self, first_name=None, last_name=None, date_of_birth=None, email=None,
                    phone=None, street_number=None, street_name=None, zip=None, match_map=None):
        """
        Find a person record.

        .. note::
            Person find must include the following minimum combinations to conduct
            a search.

            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        .. note::
            The arguments that can be passed are a selection of the possible values that
            can be used in a search. A full list of possible values can be found
            `here <https://developers.ngpvan.com/van-api#match-candidates>`_. To use these
            values, pass in a dictionary using the match_map argument.

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
            match_map: dict
                A dictionary of values to match against. Will override all
                other arguments if provided.
            fields: The fields to return. Leave as default for all available fields
        `Returns:`
            A person dict
        """

        logger.info(f'Finding {first_name} {last_name}.')

        return self._people_search(first_name, last_name, date_of_birth, email, phone,
                                   street_number, street_name, zip, match_map)

    def upsert_person(self, first_name=None, last_name=None, date_of_birth=None, email=None,
                      phone=None, street_number=None, street_name=None, zip=None, match_map=None):
        """
        Create or update a person record.

        .. note::
            Person find must include the following minimum combinations.

            - first_name, last_name, email
            - first_name, last_name, phone
            - first_name, last_name, zip5, date_of_birth
            - first_name, last_name, street_number, street_name, zip5
            - email_address

        .. note::
            The arguments that can be passed are a selection of the possible values that
            can be used in a search. A full list of possible values can be found
            `here <https://developers.ngpvan.com/van-api#match-candidates>`_. To use these
            values, pass in a dictionary using the match_map argument.

        .. warning::
            This method can only be run on MyMembers, EveryAction, MyCampaign databases.

        `Args:`
            first_name: str
                The person's first name
            last_name: str
                The person's last name
            dob: str
                ISO 8601 formatted date of birth (e.g. ``1981-02-01``)
            email: str
                The person's email address
            phone: str
                Phone number of any type (Work, Cell, Home)
            street_number: str
                Street Number
            street_name: str
                Street Name
            zip: str
                5 digit zip code
            match_map: dict
                A dictionary of values to match against. Will override all
                other arguments if provided.
        `Returns:`
            A person dict
        """

        return self._people_search(first_name, last_name, date_of_birth, email, phone,
                                   street_number, street_name, zip, match_map, create=True)

    def _people_search(self, first_name=None, last_name=None, date_of_birth=None, email=None,
                       phone=None, street_number=None, street_name=None, zip=None, match_map=None,
                       create=False):
        # Internal method to hit the people find/create endpoints

        # Ensure that the minimum combination of fields were passed
        self._valid_search(first_name, last_name, email, phone, date_of_birth, street_number,
                           street_name, zip, match_map)

        # Check to see if a match map has been provided
        if not match_map:
            json = {"firstName": first_name, "lastName": last_name}

            # Will fail if empty dicts are provided, hence needed to add if exist
            if email:
                json['emails'] = [{'email': email}]
            if phone:  # To Do: Strip out non-integers from phone
                json['phones'] = [{'phoneNumber': phone}]
            if date_of_birth:
                json['dateOfBirth'] = date_of_birth
            if zip:
                json['addresses'] = [{'zipOrPostalCode': zip}]
        else:
            json = match_map

        # Determine correct url
        url = 'people/find'
        if create:
            url = url + 'orCreate'

        return self.connection.post_request(url, json=json)

    def _valid_search(self, first_name, last_name, email, phone, dob, street_number,
                      street_name, zip, match_map):
        # Internal method to check if a search is valid

        if (None in [first_name, last_name, email] and
            None in [first_name, last_name, phone] and
            None in [first_name, last_name, zip, dob] and
            None in [first_name, last_name, street_number, street_name, zip] and
                None in [email]):

            raise ValueError("""
                             Person find must include the following minimum
                             combinations to conduct a search.
                                - first_name, last_name, email
                                - first_name, last_name, phone
                                - first_name, last_name, zip, dob
                                - first_name, last_name, street_number, street_name, zip
                                - email
                            """)

        return True

    def get_person(self, id, id_type='vanid',
                   fields=['emails', 'phones', 'custom_fields', 'external_ids', 'addresses',
                           'recorded_addresses', 'preferences', 'suppressions',
                           'reported_demographics', 'disclosure_field_values']):
        """
        Returns a single person record using their VANID or external id.

        `Args:`
            id: str
                A valid id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            fields: The fields to return. Leave as default for all available fields
        `Returns:`
            A person dict
        """

        # Change end point based on id type
        if id_type == 'vanid':
            url = f'people/{id}'
        else:
            url = f'people/{id_type}:{id}'

        fields = ','.join([json_format.arg_format(f) for f in fields])

        logger.info(f'Getting person with {id_type} of {id}')
        return self.connection.get_request(url, params={'$expand': fields})

    def apply_canvass_result(self, id, result_code_id, id_type='vanid', contact_type_id=None,
                             input_type_id=None, date_canvassed=None):
        """
        Apply a canvass result to a person. Use this end point for attempts that do not
        result in a survey response or an activist code (e.g. Not Home).

        `Args:`
            id: str
                A valid person id
            result_code_id : int
                Specifies the result code of the attempt. Valid ids can be found
                by using the :meth:`get_canvass_responses_result_codes`
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
        `Returns:`
            ``None``
        """

        logger.info(f'Applying result code {result_code_id} to {id_type} {id}.')
        self.apply_response(id, None, id_type=id_type, contact_type_id=contact_type_id,
                            input_type_id=input_type_id, date_canvassed=date_canvassed,
                            result_code_id=result_code_id)

    def toggle_volunteer_action(self, id, volunteer_activity_id, action, id_type='vanid',
                                result_code_id=None, contact_type_id=None, input_type_id=None,
                                date_canvassed=None):
        """
        Apply or remove a volunteer action to or from a person.

        `Args:`
            id: str
                A valid person id
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            volunteer_activity_id: int
                A valid volunteer activity id
            action: str
                Either 'apply' or 'remove'
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id: int
                `Optional`; A valid contact type id
            input_type_id: int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed: str
                `Optional`; ISO 8601 formatted date. Defaults to todays date

        ** NOT IMPLEMENTED **
        """

        """
        response = {"volunteerActivityId": volunteer_activity_id,
                    "action": self._action_parse(action),
                    "type": "VolunteerActivity"}

        logger.info(f'{action} volunteer activity {volunteer_activity_id} to {id_type} {id}')
        self.apply_response(id, response, id_type, contact_type_id, input_type_id, date_canvassed,
                            result_code_id)
        """

    def apply_response(self, id, response, id_type='vanid', contact_type_id=None,
                       input_type_id=None, date_canvassed=None, result_code_id=None):
        """
        Apply responses such as survey questions, activist codes, and volunteer actions
        to a person record. This method allows you apply multiple responses (e.g. two survey
        questions) at the same time. It is a low level method that requires that you
        conform to the VAN API `response object format <https://developers.ngpvan.com/van-api#people-post-people--vanid--canvassresponses>`_.

        `Args:`
            id: str
                A valid person id
            response: dict
                A list of dicts with each dict containing a valid action.
            id_type: str
                A known person identifier type available on this VAN instance
                such as ``dwid``
            result_code_id : int
                `Optional`; Specifies the result code of the response. If
                not included,responses must be specified. Conversely, if
                responses are specified, result_code_id must be null. Valid ids
                can be found by using the :meth:`get_canvass_responses_result_codes`
            contact_type_id : int
                `Optional`; A valid contact type id
            input_type_id : int
                `Optional`; Defaults to 11 (API Input)
            date_canvassed : str
                `Optional`; ISO 8601 formatted date. Defaults to todays date
            responses : list or dict
        `Returns:`
            ``True`` if successful

        .. code-block:: python

            response = [{"activistCodeId": 18917,
                         "action": "Apply",
                         "type": "ActivistCode"},
                        {"surveyQuestionId": 109149,
                         "surveyResponseId": 465468,
                         "action": "SurveyResponse"}
                        ]
            van.apply_response(5222, response)
        """  # noqa: E501,E261

        # Set url based on id_type
        if id_type == 'vanid':
            url = f"people/{id}/canvassResponses"
        else:
            url = f"people/{id_type}:{id}/canvassResponses"

        json = {"canvassContext": {
            "contactTypeId": contact_type_id,
            "inputTypeId": input_type_id,
            "dateCanvassed": date_canvassed},
            "resultCodeId": result_code_id}

        if response:
            json['responses'] = response

        if result_code_id is not None and response is not None:
            raise ValueError("Both result_code_id and responses cannot be specified.")

        if isinstance(response, dict):
            json["responses"] = [response]

        if result_code_id is not None and response is not None:
            raise ValueError(
                "Both result_code_id and responses cannot be specified.")

        return self.connection.post_request(url, json=json)

    def create_relationship(self, vanid_1, vanid_2, relationship_id):
        """
        Create a relationship between two individuals

        `Args:`
            vanid_1 : int
                The vanid of the primary individual; aka the node
            vanid_2 : int
                The vanid of the secondary individual; the spoke
            relationship_id : int
                The relationship id indicating the type of relationship
        `Returns:`
            ``None``
        """

        json = {'relationshipId': relationship_id,
                'vanId': vanid_2}

        self.connection.post_request(f"people/{vanid_1}/relationships", json=json)
        logger.info('Relationship {vanid_1} to {vanid_2} created.')



"""NGPVAN Canvass Responses Endpoints"""

from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class CanvassResponses(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_canvass_responses_contact_types(self):
        """
        Get canvass response contact types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('canvassResponses/contactTypes'))
        logger.info(f'Found {tbl.num_rows} canvass response contact types.')
        return tbl

    def get_canvass_responses_input_types(self):
        """
        Get canvass response input types.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('canvassResponses/inputTypes'))
        logger.info(f'Found {tbl.num_rows} canvass response input types.')
        return tbl

    def get_canvass_responses_result_codes(self):
        """
        Get canvass response result codes.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request('canvassResponses/resultCodes'))
        logger.info(f'Found {tbl.num_rows} canvass response result codes.')
        return tbl



import petl
import logging

logger = logging.getLogger(__name__)


class ETL(object):

    def __init__(self):

        pass

    def add_column(self, column, value=None, index=None):
        """
        Add a column to your table

        `Args:`
            column: str
                Name of column to add
            value:
                A fixed or calculated value
            index: int
                The position of the new column in the table
        `Returns:`
            `Parsons Table` and also updates self
        """

        if column in self.columns:
            raise ValueError(f"Column {column} already exists")

        self.table = self.table.addfield(column, value, index)

        return self

    def remove_column(self, *columns):
        """
        Remove a column from your table

        `Args:`
            \*columns: str
                Column names
        `Returns:`
            `Parsons Table` and also updates self
        """  # noqa: W605

        self.table = petl.cutout(self.table, *columns)

        return self

    def rename_column(self, column_name, new_column_name):
        """
        Rename a column

        `Args:`
            column_name: str
                The current column name
            new_column_name: str
                The new column name
        `Returns:`
            `Parsons Table` and also updates self
        """

        if new_column_name in self.columns:
            raise ValueError(f"Column {new_column_name} already exists")

        self.table = petl.rename(self.table, column_name, new_column_name)

        return self

    def fill_column(self, column_name, fill_value):
        """
        Fill a column in a table

        `Args:`
            column_name: str
                The column to fill
            fill_value:
                A fixed or calculated value
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.add_column(column_name + '_column_fill_temp', fill_value)
        self.remove_column(column_name)
        self.rename_column(column_name + '_column_fill_temp', column_name)

        return self

    def fillna_column(self, column_name, fill_value):
        """
        Fill None values in a column in a table

        `Args:`
            column_name: str
                The column to fill
            fill_value:
                Fixed value only
        `Returns:`
            `Parsons Table` and also updates self
        """

        self.fill_column(column_name, lambda x: x[column_name] if x[column_name] else fill_value)

        return self

    def move_column(self, column, index):
        """
        Move a column

        `Args:`
            column: str
                The column name to move
            index:
                The new index for the column
        `Returns:`
            `Parsons Table` and also updates existing object.
        """

        self.table = petl.movefield(self.table, column, index)

        return self

    def convert_column(self, *column, **kwargs):
        """
        Transform values under one or more fields via arbitrary functions, method
        invocations or dictionary translations. This leverages the petl ``convert()``
        method. Example usage can be found `here` <https://petl.readthedocs.io/en/v0.24/transform.html#petl.convert>`_

        `Args:`
            \*column: str
                A single column or multiple columns passed as a list
            \**kwargs: str, method or variable
                The update function, method, or variable to process the update
        `Returns:`
            `Parsons Table` and also updates self
        """ # noqa: E501,E261

        self.table = petl.convert(self.table, *column, **kwargs)

        return self

    def convert_columns_to_str(self):
        """
        Convenience function to convert all non-string or mixed columns in a
        Parsons table to string (e.g. for comparison)

        `Returns:`
            `Parsons Table` and also updates self
        """

        cols = self.get_columns_type_stats()

        for col in cols:
            if len(col['type']) > 1 or col['type'][0] != 'str':
                self.convert_column(col['name'], str)

        return self

    def coalesce_columns(self, dest_column, source_columns, remove_source_columns=True):
        """
        Coalesces values from one or more source columns into a destination column, by selecting
        the first non-empty value. If the destination column doesn't exist, it will be added.

        `Args:`
            dest_column: str
                Name of destination column
            source_columns: list
                List of source column names
            remove_source_columns: bool
                Whether to remove the source columns after the coalesce. If the destination
                column is also one of the source columns, it will not be removed.
        `Returns:`
            `Parsons Table` and also updates self
        """

        if dest_column in self.columns:
            def convert_fn(value, row):
                for source_col in source_columns:
                    if row.get(source_col):
                        return row[source_col]

            logger.debug(f"Coalescing {source_columns} into {dest_column}")
            self.convert_column(dest_column, convert_fn, pass_row=True)

        else:
            def add_fn(row):
                for source_col in source_columns:
                    if row.get(source_col):
                        return row[source_col]

            logger.debug(f"Creating new column {dest_column} from {source_columns}")
            self.add_column(dest_column, add_fn)

        if remove_source_columns:
            for source_col in source_columns:
                if source_col != dest_column:
                    self.remove_column(source_col)

        return self

    def map_columns(self, column_map):
        """
        Standardizes column names based on multiple possible values. This method
        is helpful when your input table might have multiple and unknown column
        names.
        `Args:`
            column_map: dict
                A dictionary of columns and possible values that map to it

        `Returns:`
            `Parsons Table` and also updates self

        .. code-block:: python

            tbl = [{fn: 'Jane'},
                   {lastname: 'Doe'},
                   {dob: '1980-01-01'}]

            column_map = {first_name: ['fn', 'first', 'firstname'],
                          last_name: ['ln', 'last', 'lastname'],
                          date_of_birth: ['dob', 'birthday']}

            tbl.map_columns(column_map)

            print (tbl)
            >> {{first_name: 'Jane', last_name: 'Doe', 'date_of_birth': '1908-01-01'}}
        """

        for c in self.columns:
            for k, v in column_map.items():
                for i in v:
                    if c == i:
                        self.rename_column(c, k)

        return self

    def get_column_types(self, column):
        """
        Return all of the Python types for values in a given column

        `Args:`
            column: str
                Name of the column to analyze
        `Returns:`
            list
                A list of Python types
        """

        return list(petl.typeset(self.table, column))

    def get_columns_type_stats(self):
        """
        Return descriptive stats for all columns

        `Returns:`
            list
                A list of dicts
        `Returns:`
            list
                A list of dicts, each containing a column 'name' and a 'type' list
        """

        return [{'name': col, 'type': self.get_column_types(col)}
                for col in self.table.columns()]

    def convert_table(self, *args):
        """
        Transform all cells in a table via arbitrary functions, method invocations or dictionary
        translations. This method is useful for cleaning fields and data hygiene functions such
        as regex. This method leverages the petl ``convert()`` method. Example usage can be
        found `here` <https://petl.readthedocs.io/en/v0.24/transform.html#petl.convert>`_.

        `Args:`
            \*args: str, method or variable
                The update function, method, or variable to process the update. Can also
        `Returns:`
            `Parsons Table` and also updates self
        """  # noqa: W605

        self.convert_column(self.columns, *args)

        return self

    def unpack_dict(self, column, keys=None, include_original=False,
                    sample_size=1000, missing=None, prepend=True,
                    prepend_value=None):
        """
        Unpack dictionary values from one column into separate columns

        `Args:`
            column: str
                The column name to unpack
            keys: list
                The dict keys in the column to unpack. If ``None`` will unpack
                all.
            include_original: boolean
                Retain original column after unpacking
            sample_size: int
                Number of rows to sample before determining columns
            missing: str
                If a value is missing, the value to fill it with
            prepend:
                Prepend the column name of the unpacked values. Useful for
                avoiding duplicate column names
            prepend_value:
                Value to prepend new columns if ``prepend=True``. If None, will
                set to column name.
        """

        if prepend:
            if prepend_value is None:
                prepend_value = column

            self.table = petl.convert(
                self.table,
                column,
                lambda v: self._prepend_dict(v, prepend_value))

        self.table = petl.unpackdict(
            self.table, column, keys=keys, includeoriginal=include_original,
            samplesize=sample_size, missing=missing)

        return self

    def unpack_list(self, column, include_original=False, missing=None, replace=False,
                    max_columns=None):
        """
        Unpack list values from one column into separate columns. Numbers the
        columns.

        .. code-block:: python

          # Begin with a list in column
          json = [{'id': '5421',
                   'name': 'Jane Green',
                   'phones': ['512-699-3334', '512-222-5478']
                  }
                 ]

          tbl = Table(json)
          print (tbl)
          >>> {'id': '5421', 'name': 'Jane Green', 'phones': ['512-699-3334', '512-222-5478']}

          tbl.unpack_list('phones', replace=True)
          print (tbl)
          >>> {'id': '5421', 'name': 'Jane Green', 'phones_0': '512-699-3334', 'phones_1': '512-222-5478'} # noqa: E501

        `Args:`
            column: str
                The column name to unpack
            include_original: boolean
                Retain original column after unpacking
            sample_size: int
                Number of rows to sample before determining columns
            missing: str
                If a value is missing, the value to fill it with
            replace: boolean
                Return new table or replace existing
            max_columns: int
                The maximum number of columns to unpack
        `Returns:`
            None
        """

        # Convert all column values to list to avoid unpack errors
        self.table = petl.convert(
            self.table, column, lambda v: [v] if not isinstance(v, list) else v
        )

        # Find the max number of values in list for all rows
        col_count = 0
        for row in self.cut(column):
            if len(row[column]) > col_count:
                col_count = len(row[column])

        # If max columns provided, set max columns
        if col_count > 0 and max_columns:
            col_count = max_columns

        # Create new column names "COL_01, COL_02"
        new_cols = []
        for i in range(col_count):
            new_cols.append(column + '_' + str(i))

        tbl = petl.unpack(self.table, column, new_cols,
                          include_original=include_original, missing=missing)

        if replace:
            self.table = tbl

        else:
            return tbl

    def unpack_nested_columns_as_rows(self, column, key='id', expand_original=False):
        """
        Unpack list or dict values from one column into separate rows.
        Not recommended for JSON columns (i.e. lists of dicts), but can handle columns
        with any mix of types. Makes use of PETL's `melt()` method.

        `Args:`
            column: str
                The column name to unpack
            key: str
                The column to use as a key when unpacking. Defaults to `id`
            expand_original: boolean or int
                If `True`: Add resulting unpacked rows (with all other columns) to original
                If `int`: Add to original unless the max added per key is above the given number
                If `False` (default): Return unpacked rows (with `key` column only) as standalone
                Removes packed list and dict rows from original either way.
        `Returns:`
            If `expand_original`, original table with packed rows replaced by unpacked rows
            Otherwise, standalone table with key column and unpacked values only
        """

        if isinstance(expand_original, int) and expand_original is not True:
            lengths = {len(row[column]) for row in self if isinstance(row[column], (dict, list))}
            max_len = sorted(lengths, reverse=True)[0]
            if max_len > expand_original:
                expand_original = False

        if expand_original:
            # Include all columns and filter out other non-dict types in table_list
            table = self
            table_list = table.select_rows(lambda row: isinstance(row[column], list))
        else:
            # Otherwise, include only key and column, but keep all non-dict types in table_list
            table = self.cut(key, column)
            table_list = table.select_rows(lambda row: not isinstance(row[column], dict))

        # All the columns other than column to ignore while melting
        ignore_cols = table.columns
        ignore_cols.remove(column)

        # Unpack lists as separate columns
        table_list.unpack_list(column, replace=True)

        # Rename the columns to retain only the number
        for col in table_list.columns:
            if f'{column}_' in col:
                table_list.rename_column(col, col.replace(f'{column}_', ""))

        # Filter dicts and unpack as separate columns
        table_dict = table.select_rows(lambda row: isinstance(row[column], dict))
        table_dict.unpack_dict(column, prepend=False)

        from parsons.etl.table import Table

        # Use melt to pivot both sets of columns into their own Tables and clean out None values
        melted_list = Table(petl.melt(table_list.table, ignore_cols))
        melted_dict = Table(petl.melt(table_dict.table, ignore_cols))

        melted_list.remove_null_rows('value')
        melted_dict.remove_null_rows('value')

        melted_list.rename_column('variable', column)
        melted_dict.rename_column('variable', column)

        # Combine the list and dict Tables
        melted_list.concat(melted_dict)

        import hashlib

        if expand_original:
            # Add unpacked rows to the original table (minus packed rows)
            orig = self.select_rows(lambda row: not isinstance(row[column], (dict, list)))
            orig.concat(melted_list)
            # Add unique id column by hashing all the other fields
            if 'uid' not in self.columns:
                orig.add_column('uid', lambda row: hashlib.md5(
                    str.encode(
                        ''.join([str(x) for x in row])
                        )
                    ).hexdigest())
                orig.move_column('uid', 0)

            # Rename value column in case this is done again to this Table
            orig.rename_column('value', f'{column}_value')

            # Keep column next to column_value
            orig.move_column(column, -1)
            output = orig
        else:
            orig = self.remove_column(column)
            # Add unique id column by hashing all the other fields
            melted_list.add_column('uid', lambda row: hashlib.md5(
                str.encode(
                    ''.join([str(x) for x in row])
                    )
                ).hexdigest())
            melted_list.move_column('uid', 0)
            output = melted_list

        self = orig
        return output

    def long_table(self, key, column, key_rename=None, retain_original=False,
                   prepend=True, prepend_value=None):
        """
        Create a new long parsons table from a column, including the foreign
        key.

        .. code-block:: python

           # Begin with nested dicts in a column
           json = [{'id': '5421',
                    'name': 'Jane Green',
                    'emails': [{'home': 'jane@gmail.com'},
                               {'work': 'jane@mywork.com'}
                              ]
                   }
                  ]
           tbl = Table(json)
           print (tbl)
           >>> {'id': '5421', 'name': 'Jane Green', 'emails': [{'home': 'jane@gmail.com'}, {'work': 'jane@mywork.com'}]} # noqa: E501
           >>> {'id': '5421', 'name': 'Jane Green', 'emails': [{'home': 'jane@gmail.com'}, {'work': 'jane@mywork.com'}]} # noqa: E501

           # Create skinny table of just the nested dicts
           email_skinny = tbl.long_table(['id'], 'emails')

           print (email_skinny)
           >>> {'id': '5421', 'emails_home': 'jane@gmail.com', 'emails_work': None}
           >>> {'id': '5421', 'emails_home': None, 'emails_work': 'jane@mywork.com'}

        `Args:`
            key: lst
                The columns to retain in the long table (e.g. foreign keys)
            column: str
                The column name to make long
            key_rename: dict
                The new name for the foreign key to better identify it. For
                example, you might want to rename ``id`` to ``person_id``.
                Ex. {'KEY_NAME': 'NEW_KEY_NAME'}
            retain_original: boolean
                Retain the original column from the source table.
            prepend:
                Prepend the column name of the unpacked values. Useful for
                avoiding duplicate column names
            prepend_value:
                Value to prepend new columns if ``prepend=True``. If None, will
                set to column name.
        `Returns:`
            Parsons Table
                The new long table
        """

        if type(key) == str:
            key = [key]

        lt = self.cut(*key, column)  # Create a table of key and column
        lt.unpack_list(column, replace=True)  # Unpack the list
        lt.table = petl.melt(lt.table, key)  # Melt into a long table
        lt = lt.cut(*key, 'value')  # Get rid of column names created in unpack
        lt.rename_column('value', column)  # Rename 'value' to old column name
        lt.remove_null_rows(column)  # Remove null values

        # If a new key name is specified, rename
        if key_rename:
            for k, v in key_rename.items():
                lt.rename_column(k, v)

        # If there is a nested dict in the column, unpack it
        if lt.num_rows > 0 and isinstance(lt.table[column][0], dict):
            lt.unpack_dict(column, prepend=prepend, prepend_value=prepend_value)

        if not retain_original:
            self.remove_column(column)

        return lt

    def cut(self, *columns):
        """
        Return a table of selection of columns

        `Args:`
            \*columns: str
                Columns in the parsons table
        `Returns:`
            A new parsons table containing the selected columnns
        """  # noqa: W605

        from parsons.etl.table import Table

        return Table(petl.cut(self.table, *columns))

    def select_rows(self, *filters):
        """
        Select specific rows from a Parsons table based on the passed
        filters.

        Example filters:
        .. code-block:: python
            tbl = Table([['foo', 'bar', 'baz'],
                         ['c', 4, 9.3],
                         ['a', 2, 88.2],
                         ['b', 1, 23.3],])

            # You can structure the filter in multiple wayss

            # Lambda Function
            tbl2 = tbl.select_rows(lambda row: row.foo == 'a' and row.baz > 88.1)
            tbl2
            >>> {foo: 'a', 'bar': 2, 'baz': 88.1}

            # Expression String
            tbl3 = tbl.select_rows("{foo} == 'a' and {baz} > 88.1")
            tbl3
            >>> {foo: 'a', 'bar': 2, 'baz': 88.1}

        `Args:`
            \*filters: function or str
        `Returns:`
            A new parsons table containing the selected rows
        """  # noqa: W605

        from parsons.etl.table import Table

        return Table(petl.select(self.table, *filters))

    def remove_null_rows(self, columns, null_value=None):
        """
        Remove rows if the values in a column are ``None``. If multiple columns
        are passed as list, it will remove all rows with null values in any
        of the passed columns.

        `Args:`
            column: str or list
                The column or columns to analyze
            null_value: int or float or str
                The null value
        `Returns:`
            ``None``
        """
        if isinstance(columns, str):
            columns = [columns]

        for col in columns:
            self.table = petl.selectisnot(self.table, col, null_value)

        return self

    def _prepend_dict(self, dict_obj, prepend):
        # Internal method to rename dict keys

        new_dict = {}

        for k, v in dict_obj.items():

            new_dict[prepend + '_' + k] = v

        return new_dict

    def stack(self, *tables, missing=None):
        """
        Stack Parsons tables on top of one another.

        Similar to ``table.concat()``, except no attempt is made to align fields from
        different tables.

        `Args:`
            tables: Parsons Table or list
                A single table, or a list of tables
            missing: bool
                The value to use when padding missing values
        `Returns:`
            ``None``
        """

        if type(tables) not in [list, tuple]:
            tables = [tables]
        petl_tables = [tbl.table for tbl in tables]

        self.table = petl.stack(self.table, *petl_tables, missing=missing)

    def concat(self, *tables, missing=None):
        """
        Concatenates one or more tables onto this one.

        Note that the tables do not need to share exactly the same fields.
        Any missing fields will be padded with None, or whatever is provided via the
        ``missing`` keyword argument.

        `Args:`
            tables: Parsons Table or list
                A single table, or a list of tables
            missing: bool
                The value to use when padding missing values
        `Returns:`
            ``None``
        """

        if type(tables) not in [list, tuple]:
            tables = [tables]
        petl_tables = [tbl.table for tbl in tables]

        self.table = petl.cat(self.table, *petl_tables, missing=missing)

    def chunk(self, rows):
        """
        Divides a Parsons table into smaller tables of a specified row count. If the table
        cannot be divided evenly, then the final table will only include the remainder.

        `Args:`
            rows: int
                The number of rows of each new Parsons table
        `Returns:`
            List of Parsons tables
        """

        from parsons import Table  # Just trying to avoid recursive imports.
        return [Table(petl.rowslice(self.table, i, i+rows)) for i in range(0, self.num_rows, rows)]

    @staticmethod
    def get_normalized_column_name(column_name):
        """
        Returns a column name with whitespace removed, non-alphanumeric characters removed, and
        everything lowercased.

        `Returns:`
            str
                Normalized column name
        """

        column_name = column_name.lower().strip()
        return ''.join(c for c in column_name if c.isalnum())

    def match_columns(self, desired_columns, fuzzy_match=True, if_extra_columns='remove',
                      if_missing_columns='add'):
        """
        Changes the column names and ordering in this Table to match a list of desired column
        names.

        `Args:`
            desired_columns: list
                Ordered list of desired column names
            fuzzy_match: bool
                Whether to normalize column names when matching against the desired column names,
                removing whitespace and non-alphanumeric characters, and lowercasing everything.
                Eg. With this flag set, "FIRST NAME" would match "first_name".
                If the Table has two columns that normalize to the same string (eg. "FIRST NAME"
                and "first_name"), the latter will be considered an extra column.
            if_extra_columns: string
                If the Table has columns that don't match any desired columns, either 'remove'
                them, 'ignore' them, or 'fail' (raising an error).
            if_missing_columns: string
                If the Table is missing some of the desired columns, either 'add' them (with a
                value of None), 'ignore' them, or 'fail' (raising an error).

        `Returns:`
            `Parsons Table` and also updates self
        """

        from parsons import Table  # Just trying to avoid recursive imports.

        normalize_fn = Table.get_normalized_column_name if fuzzy_match else (lambda s: s)

        desired_columns_normalized = {
            normalize_fn(col): col for col in desired_columns
        }

        # Check for extra columns in the Table.
        for orig_col in self.columns:
            normalized_col = normalize_fn(orig_col)
            if normalized_col not in desired_columns_normalized:
                if if_extra_columns == 'fail':
                    raise TypeError(f"Table has extra column {orig_col}")
                elif if_extra_columns == 'remove':
                    self.remove_column(orig_col)
                elif if_extra_columns != 'ignore':
                    raise TypeError(f"Invalid option {if_extra_columns} for "
                                    "argument `if_extra_columns`")
            else:
                # We matched a desired column. Remove it from our list, so if there is another
                # column in our Table that has the same normalized name, we consider it an
                # extra column.
                desired_columns_normalized.pop(normalized_col)

        # Regenerate the desired columns normalized dict, since we removed any matches
        # from it above.
        # Note we reverse the desired columns here, to make reordering easier later.
        # Cast desired_columns to a list in case someone gave us a dict or other iterable
        # that can't be reversed.
        desired_columns_normalized = {
            normalize_fn(col): col for col in reversed(list(desired_columns))
        }

        tbl_columns_normalized = {
            normalize_fn(col): col for col in self.columns
        }

        # Check for missing columns
        for normalized_col, orig_col in desired_columns_normalized.items():
            if normalized_col not in tbl_columns_normalized:
                if if_missing_columns == 'fail':
                    raise TypeError(f"Table is missing column {orig_col}")
                elif if_missing_columns == 'add':
                    self.add_column(orig_col)
                    tbl_columns_normalized[normalized_col] = orig_col
                elif if_missing_columns != 'ignore':
                    raise TypeError(f"Invalid option {if_missing_columns} for "
                                    "argument `if_missing_columns`")

        # Change column ordering and names to match the desired columns
        for desired_normalized_col, desired_orig_col in desired_columns_normalized.items():
            # Note that we ignore any desired columns still not in the Table, given
            # that we already checked what the caller wanted to do above.
            if desired_normalized_col in tbl_columns_normalized:
                tbl_orig_col = tbl_columns_normalized[desired_normalized_col]
                if tbl_orig_col != desired_orig_col:
                    self.rename_column(tbl_orig_col, desired_orig_col)
                self.move_column(desired_orig_col, 0)

        return self

    def reduce_rows(self, columns, reduce_func, headers, presorted=False,
                    **kwargs):
        """
        Group rows by a column or columns, then reduce the groups to a single row.

        Based on the `rowreduce` petl function - https://petl.readthedocs.io/en/stable/transform.html#petl.transform.reductions.rowreduce # noqa

        For example, the output from the query to get a table's definition is
        returned as one component per row. The `reduce_rows` method can be used
        to reduce all those to a single row containg the entire query.

        .. code-block:: python

            >>> ddl = rs.query(sql_to_get_table_ddl)
            >>> ddl.table

            +--------------+--------------+----------------------------------------------------+
            | schemaname   | tablename    | ddl                                                |
            +==============+==============+====================================================+
            | 'db_scratch' | 'state_fips' | '--DROP TABLE db_scratch.state_fips;'              |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | 'CREATE TABLE IF NOT EXISTS db_scratch.state_fips' |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '('                                                |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '\tstate VARCHAR(1024)   ENCODE RAW'              |
            +--------------+--------------+----------------------------------------------------+
            | 'db_scratch' | 'state_fips' | '\t,stusab VARCHAR(1024)   ENCODE RAW'            |
            +--------------+--------------+----------------------------------------------------+

            >>> reducer_fn = lambda columns, rows: [
            ...     f"{columns[0]}.{columns[1]}",
            ...     '\n'.join([row[2] for row in rows])]
            >>> ddl.reduce_rows(
            ...     ['schemaname', 'tablename'],
            ...     reducer_fn,
            ...     ['tablename', 'ddl'],
            ...     presorted=True)
            >>> ddl.table

            +-------------------------+------------------------------------------------------------------+
            | tablename               | ddl                                                              |
            +=========================+==================================================================+
            | 'db_scratch.state_fips' | '--DROP TABLE db_scratch.state_fips;\nCREATE TABLE IF NOT EXISTS |
            |                         | db_scratch.state_fips\n(\n\tstate VARCHAR(1024)   ENCODE RAW\n\t |
            |                         | ,db_scratch.state_fips\n(\n\tstate VARCHAR(1024)   ENCODE RAW    |
            |                         | \n\t,stusab VARCHAR(1024)   ENCODE RAW\n\t,state_name            |
            |                         | VARCHAR(1024)   ENCODE RAW\n\t,statens VARCHAR(1024)   ENCODE    |
            |                         | RAW\n)\nDISTSTYLE EVEN\n;'                                       |
            +-------------------------+------------------------------------------------------------------+

        `Args:`
            columns: list
                The column(s) by which to group the rows.
            reduce_func: fun
                The function by which to reduce the rows. Should take the 2
                arguments, the columns list and the rows list and return a list.
                `reducer(columns: list, rows: list) -> list;`
            headers: list
                The list of headers for modified table. The length of `headers`
                should match the length of the list returned by the reduce
                function.
            presorted: bool
                If false, the row will be sorted.
        `Returns:`
            `Parsons Table` and also updates self

        """

        self.table = petl.rowreduce(
            self.table,
            columns,
            reduce_func,
            header=headers,
            presorted=presorted,
            **kwargs)

        return self



# Provide shortcuts to importing Parsons submodules
# Eg. This allows for: `from parsons import VAN`
from parsons.ngpvan.van import VAN
from parsons.targetsmart.targetsmart_api import TargetSmartAPI
from parsons.targetsmart.targetsmart_automation import TargetSmartAutomation
from parsons.mobile_commons.mobile_commons import MobileCommons
from parsons.databases.redshift.redshift import Redshift
from parsons.aws.s3 import S3
from parsons.civis.civisclient import CivisClient
from parsons.etl.table import Table
from parsons.notifications.gmail import Gmail
from parsons.google.google_civic import GoogleCivic
from parsons.google.google_sheets import GoogleSheets
from parsons.google.google_cloud_storage import GoogleCloudStorage
from parsons.google.google_bigquery import GoogleBigQuery
from parsons.phone2action.p2a import Phone2Action
from parsons.mobilize_america.ma import MobilizeAmerica
from parsons.facebook_ads.facebook_ads import FacebookAds
from parsons.notifications.slack import Slack
from parsons.turbovote.turbovote import TurboVote
from parsons.sftp.sftp import SFTP
from parsons.action_kit.action_kit import ActionKit
from parsons.geocode.census_geocoder import CensusGeocoder
from parsons.airtable.airtable import Airtable
from parsons.copper.copper import Copper
from parsons.crowdtangle.crowdtangle import CrowdTangle
from parsons.hustle.hustle import Hustle
from parsons.twilio.twilio import Twilio

__all__ = [
    'VAN',
    'TargetSmartAPI',
    'TargetSmartAutomation',
    'MobileCommons',
    'Redshift',
    'S3',
    'CivisClient',
    'Table',
    'Gmail',
    'GoogleCivic',
    'GoogleCloudStorage',
    'GoogleBigQuery',
    'GoogleSheets',
    'Phone2Action',
    'MobilizeAmerica',
    'FacebookAds',
    'Slack',
    'TurboVote',
    'SFTP',
    'ActionKit',
    'CensusGeocoder',
    'Airtable',
    'Copper',
    'CrowdTangle',
    'Hustle',
    'Twilio'
    ]

# Define the default logging config for Parsons and its submodules. For now the
# logger gets a StreamHandler by default. At some point a NullHandler may be more
# appropriate, so the end user must decide on logging behavior.
import logging
import os
logger = logging.getLogger(__name__)
_handler = logging.StreamHandler()
_formatter = logging.Formatter('%(module)s %(levelname)s %(message)s')
_handler.setFormatter(_formatter)
logger.addHandler(_handler)
if os.environ.get('TESTING'):
    # Log less stuff in automated tests
    logger.setLevel('WARNING')
else:
    logger.setLevel('INFO')



from parsons import Gmail
import json
import os
import requests_mock
import unittest
import shutil
import base64
import email


_dir = os.path.dirname(__file__)


class TestGmail(unittest.TestCase):

    @requests_mock.Mocker()
    def setUp(self, m):
        self.tmp_folder = "tmp/"
        self.credentials_file = f"{self.tmp_folder}/credentials.json"
        self.token_file = f"{self.tmp_folder}/token.json"

        os.mkdir(self.tmp_folder)

        with open(self.credentials_file, 'w') as f:
            f.write(json.dumps({
              "installed": {
                "client_id": "someclientid.apps.googleusercontent.com",
                "project_id": "some-project-id-12345",
                "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                "token_uri": "https://www.googleapis.com/oauth2/v3/token",
                "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                "client_secret": "someclientsecret",
                "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob", "http://localhost"]
              }
            }))

        with open(self.token_file, 'w') as f:
            f.write(json.dumps({
              "access_token": "someaccesstoken",
              "client_id": "some-client-id.apps.googleusercontent.com",
              "client_secret": "someclientsecret",
              "refresh_token": "1/refreshrate",
              "token_expiry": "2030-02-20T23:28:09Z",
              "token_uri": "https://www.googleapis.com/oauth2/v3/token",
              "user_agent": None,
              "revoke_uri": "https://oauth2.googleapis.com/revoke",
              "id_token": None,
              "id_token_jwt": None,
              "token_response": {
                "access_token": "someaccesstoken",
                "expires_in": 3600000,
                "scope": "https://www.googleapis.com/auth/gmail.send",
                "token_type": "Bearer"
              },
              "scopes": ["https://www.googleapis.com/auth/gmail.send"],
              "token_info_uri": "https://oauth2.googleapis.com/tokeninfo",
              "invalid": False,
              "_class": "OAuth2Credentials",
              "_module": "oauth2client.client"
            }))

        self.gmail = Gmail(self.credentials_file, self.token_file)

    def tearDown(self):
        # Delete tmp folder and files
        shutil.rmtree(self.tmp_folder)

    def test_create_message_simple(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email"
        message_text = "The is the message text of the email"

        raw = self.gmail.create_message_simple(
            sender, to, subject, message_text)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'text/plain; charset="us-ascii"'),
            ('MIME-Version', '1.0'),
            ('Content-Transfer-Encoding', '7bit'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # Check the metadata
        self.assertListEqual(decoded.items(), expected_items)

        # Check the message
        self.assertEqual(decoded.get_payload(), message_text)

        # Check the number of parts
        expected_parts = 1
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_html(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test html email"
        message_text = "The is the message text of the email"
        message_html = "<p>This is the html message part of the email</p>"

        raw = self.gmail.create_message_html(
            sender, to, subject, message_text, message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('subject', subject),
            ('from', sender),
            ('to', to)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        # Check the number of parts
        expected_parts = 3
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_html_no_text(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test html email"
        message_html = "<p>This is the html message part of the email</p>"

        raw = self.gmail.create_message_html(
            sender, to, subject, '', message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('subject', subject),
            ('from', sender),
            ('to', to)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_html)

        # Check the number of parts
        expected_parts = 2
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.txt']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_txt.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), 'text/plain')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_jpeg(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.jpeg']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_jpeg.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        expected_id = f"<{attachments[0].split('/')[-1]}>"
        self.assertEqual(parts[2].get('Content-ID'), expected_id)
        self.assertEqual(parts[2].get_content_type(), 'image/jpeg')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_m4a(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.m4a']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_m4a.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_maintype(), 'audio')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_mp3(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.mp3']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_mp3.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), 'audio/mpeg')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_mp4(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.mp4']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_mp4.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), 'video/mp4')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test_create_message_attachments_pdf(self):
        sender = "Sender <sender@email.com>"
        to = "Recepient <recepient@email.com>"
        subject = "This is a test email with attachements"
        message_text = "The is the message text of the email with attachments"
        message_html = ("<p>This is the html message part of the email "
                        "with attachments</p>")
        attachments = [f'{_dir}/assets/loremipsum.pdf']

        raw = self.gmail.create_message_attachments(
            sender, to, subject, message_text, attachments,
            message_html=message_html)

        decoded = email.message_from_bytes(
            base64.urlsafe_b64decode(bytes(raw['raw'], 'utf-8')))

        expected_items = [
            ('Content-Type', 'multipart/alternative;\n boundary='),
            ('MIME-Version', '1.0'),
            ('to', to),
            ('from', sender),
            ('subject', subject)]

        # The boundary id changes everytime. Replace it with the beginnig to
        # avoid failures
        updated_items = []
        for i in decoded.items():
            if 'Content-Type' in i[0] and 'multipart/alternative;\n boundary=' in i[1]:  # noqa
                updated_items.append(
                    ('Content-Type', 'multipart/alternative;\n boundary='))
            else:
                updated_items.append((i[0], i[1]))

        # Check the metadata
        self.assertListEqual(updated_items, expected_items)

        # Check the message
        # The first part is just a container for the text and html parts
        parts = decoded.get_payload()

        self.assertEqual(parts[0].get_payload(), message_text)
        self.assertEqual(parts[1].get_payload(), message_html)

        with open(f'{_dir}/assets/loremipsum_b64_pdf.txt', 'r') as f:
            b64_txt = f.read()
        self.assertEqual(parts[2].get_payload(), b64_txt)

        self.assertEqual(parts[2].get_content_type(), 'application/pdf')

        # Check the number of parts
        expected_parts = 4
        self.assertEqual(sum([1 for i in decoded.walk()]), expected_parts)

    def test__validate_email_string(self):
        emails = [
            {"email": "Sender <sender@email.com>", "expected": True},
            {"email": "sender@email.com", "expected": True},
            {"email": "<sender@email.com>", "expected": True},
            {"email": "Sender sender@email.com", "expected": False},
            {"email": "Sender <sender2email.com>", "expected": False},
            {"email": "Sender <sender@email,com>", "expected": True},
            {"email": "Sender <sender+alias@email,com>", "expected": True}
        ]

        for e in emails:
            if e['expected']:
                self.assertTrue(self.gmail._validate_email_string(e['email']))
            else:
                self.assertRaises(
                    ValueError, self.gmail._validate_email_string, e['email'])

    # TODO test sending emails


if __name__ == "__main__":
    unittest.main()



import pytest
import os
from parsons import SFTP, Table
from parsons.utilities import files
from test.utils import mark_live_test, assert_matching_tables
from test.fixtures import simple_table, simple_csv_path, simple_compressed_csv_path

#
# Fixtures and constants
#

REMOTE_DIR = 'parsons-test'
REMOTE_CSV = 'test.csv'
REMOTE_CSV_PATH = f'{REMOTE_DIR}/{REMOTE_CSV}'
REMOTE_COMPRESSED_CSV = 'test.csv.gz'
REMOTE_COMPRESSED_CSV_PATH = f'{REMOTE_DIR}/{REMOTE_COMPRESSED_CSV}'

@pytest.fixture
def live_sftp(simple_table, simple_csv_path, simple_compressed_csv_path):
    # Generate a live SFTP connection based on these env vars
    host = os.environ['SFTP_HOST']
    username = os.environ['SFTP_USERNAME']
    password = os.environ['SFTP_PASSWORD']

    sftp = SFTP(host, username, password)

    # Add a test directory and test files
    sftp.make_directory(REMOTE_DIR)
    sftp.put_file(simple_csv_path, REMOTE_CSV_PATH)
    sftp.put_file(simple_compressed_csv_path, REMOTE_COMPRESSED_CSV_PATH)

    yield sftp

    # Cleanup after test
    sftp.remove_file(REMOTE_CSV_PATH)
    sftp.remove_file(REMOTE_COMPRESSED_CSV_PATH)
    sftp.remove_directory(REMOTE_DIR)

#
# Tests
#

def test_credential_validation():
    with pytest.raises(ValueError):
        SFTP(host=None, username=None, password=None)

    with pytest.raises(ValueError):
        SFTP(host=None, username='sam', password='abc123')

@mark_live_test
def test_list_non_existent_directory(live_sftp):
    with pytest.raises(FileNotFoundError):
        file_list = live_sftp.list_directory('abc123')

@mark_live_test
def test_list_directory_with_files(live_sftp):
    file_list = live_sftp.list_directory(REMOTE_DIR)
    assert file_list == [REMOTE_COMPRESSED_CSV, REMOTE_CSV]

@mark_live_test
def test_get_non_existent_file(live_sftp):
    with pytest.raises(FileNotFoundError):
        live_sftp.get_file('abc123')

# Helper function
def assert_file_matches_table(local_path, table):
    downloaded_tbl = Table.from_csv(local_path)
    assert_matching_tables(table, downloaded_tbl)

@mark_live_test
def test_get_file(live_sftp, simple_table):
    local_path = files.create_temp_file()
    live_sftp.get_file(REMOTE_CSV_PATH, local_path=local_path)
    assert_file_matches_table(local_path, simple_table)

@mark_live_test
def test_get_temp_file(live_sftp, simple_table):
    local_path = live_sftp.get_file(REMOTE_CSV_PATH)
    assert_file_matches_table(local_path, simple_table)

@mark_live_test
@pytest.mark.parametrize('compression', [None, 'gzip'])
def test_table_to_sftp_csv(live_sftp, simple_table, compression):
    host = os.environ['SFTP_HOST']
    username = os.environ['SFTP_USERNAME']
    password = os.environ['SFTP_PASSWORD']
    remote_path = f'{REMOTE_DIR}/test_to_sftp.csv'
    if compression == 'gzip':
        remote_path += '.gz'
    simple_table.to_sftp_csv(remote_path, host, username, password, compression=compression)

    local_path = live_sftp.get_file(remote_path)
    assert_file_matches_table(local_path, simple_table)

    # Cleanup
    live_sftp.remove_file(remote_path)


# Stuff that is tested by the live_sftp fixture, so no need to test explicitly:
# test_make_directory
# test_put_file
# test_put_file_compressed
# def test_remove_file
# def test_remove_directory
